#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2013 Gary Wright wrightsolutions.co.uk/contact
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# Tested against fabric 1.4.3
# http://ftp.uk.debian.org/debian/pool/main/f/fabric/fabric_1.4.3-1.debian.tar.gz

from __future__ import with_statement
from fabric.api import abort, cd, env, get, hide, put, run, settings, sudo
from fabric.contrib.console import confirm
from fabric.contrib.files import exists,append
from os import getuid,getgid,path
from os import linesep as os_linesep
from string import ascii_lowercase,digits,printable,punctuation,whitespace
import re
from subprocess import Popen, PIPE
import shlex
import time
from unicodedata import name as unicode_name
from collections import OrderedDict
from copy import deepcopy
from hashlib import md5

""" res.succeeded is preferred for newer fabric, but not res.failed
is used where backward compatibility is preferable. """

CMD_CAT = "/bin/cat "
CMD_TAC = "/usr/bin/tac "
CMD_DIFF="/usr/bin/diff "
CMD_DIFF_VERSION="/usr/bin/diff --version;"
CMD_DIFF_UNIFIED_ZERO="/usr/bin/diff -U 0 "
CMD_DIFF_SIDEBYSIDE_NOCOMMON="/usr/bin/diff --side-by-side --suppress-common-lines "
CMD_EGREP = "/bin/egrep "
CMD_EGREPI = "/bin/egrep -I "
CMD_FGREP = "/bin/fgrep "
CMD_FGREP_H_V = "/bin/fgrep -H -v "
CMD_FGREP_H_V_Z = "/bin/fgrep -H -v 'zZzZz!zZzZz' "
CMD_WGET = "/usr/bin/wget --no-check-certificate "
CMD_WGET1 = "/usr/bin/wget --no-check-certificate --timeout=1 "
CMD_GPG_RECV = "gpg --keyserver keys.gnupg.net --recv-keys "
CMD_GPG_VERIFY = "gpg --verify --no-tty --batch "
CMD_GPG_LIST = "gpg --list-key --no-tty --batch "
CMD_KERN_ALL = "/bin/uname -a "
CMD_KERN_RELEASE = "/bin/uname -r "
CMD_LOCKFILE_CREATE = "/usr/bin/lockfile-create "
CMD_LSOF_INTERNET = "/usr/bin/lsof -i "
CMD_MERCURIAL_CLONE = "/usr/bin/hg clone "
CMD_GIT_CLONE = "/usr/bin/git clone "
CMD_ALTERNATIVES_DISPLAY = "/usr/sbin/update-alternatives --display "
CMD_FIND = "/usr/bin/find "
CMD_WHICH = "/usr/bin/which "
CMD_SED = "/bin/sed "
CMD_SEDE = "/bin/sed -e "
CMD_SEDF = "/bin/sed -f "
CMD_SEDN = "/bin/sed -n "
CMD_SEDNF = "/bin/sed -n -f "
CMD_SEDNR = "/bin/sed -n -r "
CMD_SEDR = "/bin/sed -r "
CMD_SED16 = "/bin/sed -n 'p;p;p;p;p;p;p;p;p;p;p;p;p;p;p;p' "
CMD_MD5SUM = "/usr/bin/md5sum "
CMD_MD5SUM_TEXT = "/usr/bin/md5sum --text "
CMD_SHA256 = "/usr/bin/sha256sum "
CMD_NPROC = "/usr/bin/nproc "
CMD_FREE = "/usr/bin/free "
CMD_FREE_M_T = "/usr/bin/free -m -t "

_list_dot_d_array = []

# DIRSTEM_DATA_SHORT = '~/'
DIRSTEM_DATA_SHORT = '/var/local'

RE_FIRST_ALPHA = re.compile('[a-z]+')
RE_VAR_LIB_ALPHA = re.compile('/var/lib/[a-z][a-z]+')
RE_OPT_LOCAL_ALPHA = re.compile('/opt/local/[a-z][a-z]+')
RE_VAR_LOCAL_ALPHA = re.compile('/var/local/[a-z][a-z]+')
RE_STATUS_INSTALLED = re.compile('Status:.*installed')

REGEX_OPTS_EGREP = '-regextype posix-egrep -regex'
REGEX_OPTS_EMACS = '-regextype emacs -regex'
EXEC_TAIL = '{} \;'

NUM_AS_WORDS_TO_TWENTY=['zero','one','two','three','four','five','six','seven','eight','nine','ten']
NUM_AS_WORDS_TO_TWENTY.extend(['eleven','twelve','thirteen','fourteen','fifteen'])
NUM_AS_WORDS_TO_TWENTY.extend(['sixteen','seventeen','eighteen','nineteen','twenty'])

SET_PRINTABLE = set(printable)
SET_PUNCTUATION = set(punctuation)
SET_LOWER_PLUS_DIGITS = set(ascii_lowercase).union(digits)
SET_WHITESPACE = set(whitespace)
SET_COLON_SPACE_EQUALS = set([chr(58),chr(32),chr(61)])

ISOCOMPACT_STRFTIME='%Y%m%dT%H%M%S'

DEBIAN_PLUS_DERIVATIVES1=['debian','trisquel']
DEBIAN_PLUS_DERIVATIVES2=DEBIAN_PLUS_DERIVATIVES1[:]
DEBIAN_PLUS_DERIVATIVES2.extend(['gnewsense','mint','ubuntu'])
""" Lists above could be extended by using the huge svg graph 'DebianFamilyTree' on public internet.
That DebianFamilyTree has two large sections hence the 1 and 2 suffixes to our lists """
RPM_BASED1=['redhat','centos']
RPM_BASED2=RPM_BASED1[:]
RPM_BASED2.extend(['fedora'])


def run_or_sudo(run_cmd,run_always=False,quiet=False):
    """ run_always flag set True means
        that sudo will never be used. """
    if env.user == 'root' or run_always:
        if quiet:
            with hide('output'):
                res = run(run_cmd)
        else:
            res = run(run_cmd)
    elif quiet:
        with hide('output'):
            res = sudo(run_cmd)
    else:
        res = sudo(run_cmd)
    return res


def print_success_and_failed(fab_object,dirpath=None):
    succ_plus_failed = -999
    try:
        succ_plus_failed = int(fab_object.succeeded) + \
            int(fab_object.failed)
    except Exception:
        succ_plus_failed = -1

    if dirpath:
        print "{0} success and failed: {1} {2}".format(
            dirpath,fab_object.succeeded,fab_object.failed)
    else:
        print "{0} success and failed: {1} {2}".format(
            dirpath,fab_object.succeeded,fab_object.failed)
    return succ_plus_failed


def md5_from_iterable(iterable,md5sum_default=None):
    md5sum = md5sum_default
    mdf = md5()
    idx = 0
    for item_string in iterable:
        mdf.update(item_string)
        idx += 1
    if idx < 1:
        md5sum = md5sum_default
    else:
        md5sum = mdf.hexdigest()
    return md5sum


def md5dict_from_iterable(iterable,linesep=None):
    """ Testing helper. Returns an OrderedDict of the following:
    individual line md5s and counts (an md5 for each item in the iterable)
    followed by the counts which all have key values beginning 'count_'
    space then linefeed ' \n' is d784fa8b6d98d27699781bd9a7cf19f0
    bare linefeed is 68b329da9893e34099c7d8ad5cb9c940
    space alone is unlikely but anyway 7215ee9c7d9dc229d2921a40e899ec5f
    """
    md5dict_counts_zero = OrderedDict()
    md5dict_counts_zero['count_space_alone'] = 0
    md5dict_counts_zero['count_space_linesep'] = 0
    md5dict_counts_zero['count_spaces_linesep'] = 0
    md5dict_counts_zero['count_linesep_alone'] = 0
    md5dict_counts_zero['count_other'] = 0
    md5dict = md5dict_counts_zero.copy()
    if linesep is None:
        linesep = os_linesep
    SPACE_LINESEP = "{0}{1}".format(chr(32),linesep)
    idx = 0
    for item_string in iterable:
        try:
            mdf = md5(item_string).hexdigest()
            md5dict[idx] = mdf
            if item_string == SPACE_LINESEP:
                md5dict['count_space_linesep'] += 1
            elif item_string == chr(32):
                md5dict['count_space_alone'] += 1
            elif item_string == linesep:
                md5dict['count_linesep_alone'] += 1
            elif item_string.startswith(chr(32)) and item_string.lstrip() == linesep:
                md5dict['count_spaces_linesep'] += 1
            else:
                md5dict['count_other'] += 1
        except AttributeError:
            """ Zero all the counts on error """
            md5dict = md5dict_counts_zero.copy()
            break
        except Exception:
            """ Zero all the counts on error """
            md5dict = md5dict_counts_zero.copy()
            break
        idx += 1
    return md5dict


def md5list_from_iterable(iterable):
    """ Testing helper. Returns a list of md5s 
    (an md5 for each item in the iterable)
    space then linefeed ' \n' is d784fa8b6d98d27699781bd9a7cf19f0
    bare linefeed is 68b329da9893e34099c7d8ad5cb9c940
    space alone is unlikely but anyway 7215ee9c7d9dc229d2921a40e899ec5f
    """
    md5list = []
    idx = 0
    for item_string in iterable:
        try:
            md5list.append(md5(item_string).hexdigest())
        except AttributeError:
            md5list = []
            break
        except Exception:
            md5list = []
            break
        idx += 1
    return md5list


def iterable1iterable2compare(iterable1,iterable2,summary_numbers_only=True,
                              missing_count_exception=True):
    """ Compare values of keys beginning 'count_' in two lists / iterables.
    Useful for detecting small changes in formatting / spacing / line endings
    between a file and its sed result output.
    Designed for dealing with counts so non-integers should error.
    """
    iterable1dict = md5dict_from_iterable(iterable1)
    iterable2dict = md5dict_from_iterable(iterable2)

    counts1tuples = [(k, v) for k, v in iterable1dict.iteritems() if str(k).startswith('count_')]
    counts2tuples = [(k, v) for k, v in iterable2dict.iteritems() if str(k).startswith('count_')]
    counts_dict1 = OrderedDict(counts1tuples)
    counts_dict2 = OrderedDict(counts2tuples)
    counts_dict_combined = counts_dict1.copy()
    counts_dict_combined.update(counts_dict2)
    compared_dict = OrderedDict()
    for count_key in counts_dict_combined.iterkeys():
        if count_key in counts_dict1 and count_key in counts_dict2:
            count1 = 0
            count2 = 0
            try:
                count1 = int(counts_dict1[count_key])
                count2 = int(counts_dict2[count_key])
            except ValueError:
                raise Exception("...2compare() only possible for integers")
            except Exception:
                raise
            count_diff = count1-count2
            compared_dict[count_key] = count_diff
        elif missing_count_exception:
            raise Exception("{0} key not found in both iterable dicts!".format(count_key))
        else:
            pass

    compared_as_tuple = tuple(compared_dict.itervalues())

    any_result = any(compared_as_tuple)
    text_as_list = ['Exactly same. (All counts equal)']
    if any_result:
        text_as_list = ['Different. (Some difference in counts)']
    if summary_numbers_only is True:
        pass
    else:
        if 'count_other' in compared_dict:
            count_other = compared_dict['count_other']
            if count_other == 0:
                other = counts_dict1['count_other']
                """ Above assignement of 'other' could have been sourced from
                counts_dict1 or counts_dict2 as we have just been told they are same
                """
                text_as_list.append("count_other={0} (same in both)".format(other))
            else:
                text_as_list.append("count_other={0} (differs)".format(count_other))
        sum_result = sum(compared_as_tuple)
        if sum_result == 0:
            text_as_list.append('Tally may have been shifted')
            text_as_list.append('between counts as sum()==0 ')
    return (text_as_list,compared_as_tuple)

def kern64():
    with settings(warn_only=True):
        kern64line = "{0} | {1} 'x86_64'".format(CMD_KERN_ALL,CMD_FGREP)
        kern64res = run(kern64line)
    if kern64res.failed:
        return False
    return True


def splitlines_retaining(lines_unsplit,linesep=None):
    """ If your lines are line separated with something other than
    the value of os.linesep then use the linesep parameter to supply that.
    This will in some cases return a different result than
    plain splitlines(True) due to folding / loss of any blank lines.
    """
    lines = []
    if linesep is None:
        linesep = os_linesep
        for line in lines_unsplit.split(linesep):
            lines.append("{0}{1}".format(line,linesep))
    return lines


def run_res_lines_stripped(fabric_res,linesep=None):
    """ Two conditions to pass before any lines will be added to list being returned:
    (1) fabric result object must have attribute succeeded set
    (2) fabric result object must have return_code set zero
    If your results are line separated with something other than
    the value of os.linesep then use the linesep parameter to supply that.
    """
    lines_stripped_as_list = []
    if fabric_res.succeeded:
        if fabric_res.return_code == 0:
            if linesep is None:
                linesep = os_linesep
            for line_unstripped in fabric_res.split(linesep):
                lines_stripped_as_list.append(line_unstripped.strip())
    return lines_stripped_as_list


def unicode_name_index(uname,unicode_input):
    idx = -1
    for idx,uni in enumerate(unicode_input):
        if unicode_name(uni) == uname:
            uname_index = idx
    return idx


def local_popen_wrapper(local_cmd,workdir=None,
                        linesep=None,allow_root=False,buffer_size=None):
    """ Popen for local commands where preservation of exact contents of
    stdout is important. Losing whitespace is not really an issue for most
    automation / sysadmin work, but if you want a faithful byte by byte
    representation of the output from a sed / awk then this function may help.
    Small amounts of effort have been put into the following:
    (.) Ensuring any root invocation really is what the user wants.
    (.) Simple blocks (easily bypassed) for move and remove
    Blocks on mv and rm are really just a convenient warning and the message
    suggests that user may want to use fabric own local() command.
    """
    if getuid() == 0:
        if allow_root == True:
            pass
        else:
            raise Exception("Set allow_root=True for root user invocation")
    if workdir is None:
        pass
    elif set(workdir).isdisjoint(SET_WHITESPACE):
        """ Catch misuse where linesep is given as workdir.
        Here we pass as seems usable """
        pass
    else:
        raise Exception("Whitespace issue workdir={0}".format(workdir))
    if local_cmd.lstrip().startswith('/bin/mv') or local_cmd.startswith('/bin/rm'):
        raise Exception("Move and Remove are blocked. Use fabric own local() facility")
    elif 'passwd' in local_cmd or 'rm -f' in local_cmd or 'rm -fR' in local_cmd:
        raise Exception("Blocked command. Use fabric own local() facility")
    else:
        pass
    if buffer_size is None or buffer_size < 8192:
        buffer_size = 16384
    #print shlex.split(local_cmd)
    ploc = Popen(shlex.split(local_cmd),bufsize=buffer_size,
                 stdout=PIPE,shell=False,cwd=workdir)
    ploc_stdout, ploc_stderr = ploc.communicate()
    ploc_rc = ploc.returncode
    #print ploc_rc,ploc_stdout
    return (ploc_rc,ploc_stdout,ploc_stderr)


def local_popen_split_retaining(local_cmd,workdir=None,
                                linesep=None,allow_root=False,buffer_size=None):
    ploc_rc, ploc_stdout, ploc_stderr = local_popen_wrapper(local_cmd,workdir,linesep,
                                                            allow_root,buffer_size)
    lines = []
    if ploc_rc == 0:
        lines = ploc_stdout.splitlines(True)
    #print ploc_rc
    return (ploc_rc,lines)


def add_nonlabel_not_already(target_dict,label,line):
    added_count = 0
    if label not in target_dict:
        target_dict[label]=line.split(chr(58))[-1].strip()
        added_count += 1
    return added_count


def cpu_metrics(post_formatted=True,include_missing=True,flag_per_line=False,quiet=False,
                flag_list=['lm','vmx','svm','sse2','sse4_1','sse4_2','avx','xop','fma4']):
    """ Returns output as a list or empyty list if failure """
    with settings(warn_only=True):
        cpuinfo_line = "{0} -i 'mhz|mips|qemu' /proc/cpuinfo".format(CMD_EGREP)
        if flag_per_line:
            cpuinfo_line = "{0} -i 'mhz|mips|qemu|flags' /proc/cpuinfo".format(CMD_EGREP)
        cpuinfo_res = run_or_sudo(cpuinfo_line,False,quiet)
    cpuinfo_lines = run_res_lines_stripped(cpuinfo_res)
    cpuinfo_dict = {}
    if post_formatted:
        """ Design choice made in this section is that hosts with multiple
        cpu allocation should be processed in a way that does not repeat
        the specifics. So have used add_nonlabel_not_already() to help achieve this
        """
        for line_tabbed in cpuinfo_lines[:]:
            line = line_tabbed.replace(chr(9),'')
            if 'bogomips' in line:
                add_nonlabel_not_already(cpuinfo_dict,'bogomips',line)
            elif line.startswith('model') and 'qemu' in line.lower():
                add_nonlabel_not_already(cpuinfo_dict,'qemu',line)
            elif line.startswith('cpu') and 'mhz' in line.lower():
                add_nonlabel_not_already(cpuinfo_dict,'mhz',line)
            elif line.startswith('flags'):
                for flag in line.split(chr(58))[-1].split(chr(32)):
                    if flag in flag_list:
                        label = "flag_{0}".format(flag)
                        add_nonlabel_not_already(cpuinfo_dict,label,'yes')
            else:
                pass
        if include_missing:
            if 'qemu' not in cpuinfo_dict:
                cpuinfo_dict['qemu']='no'
    if len(cpuinfo_dict) > 0:
        cpuinfo_lines = []
        try:
            for label in sorted(cpuinfo_dict.iterkeys()):
                cpuinfo_lines.append("{0}:{1}".format(label,cpuinfo_dict[label]))
        except ValueError,e:
            raise
    return cpuinfo_lines
        

def cpu64():
    flag64 = False
    cpuinfo_lines = cpu_metrics(True,False,True,True)
    for line in cpuinfo_lines:
        if line.startswith('flag_lm'):
            flag64 = True
    return flag64


def pretty_name():
    pretty_tuple = pretty_name_tuple()
    return pretty_tuple[0]


def pretty_name_tuple():
    pretty_name_id = 'unknown'
    pretty_name = 'Unknown!'
    pretty_name_word = 'zero'
    with settings(warn_only=True):
        pretty_cat = "{0} /etc/system-release".format(CMD_CAT)
        pretty_res = run(pretty_cat)
        if pretty_res.succeeded:
            """ Both types of quotes stripped below """
            pretty_name = pretty_res.strip('"\\\'')

    if pretty_name.endswith(chr(33)):
        """ Exclamation mark (!) is chr(33). We are probably not Rpm based """
        pretty_grep = "{0} '^PRETTY_NAME' /etc/os-release".format(CMD_EGREP)
        pretty_res = run(pretty_grep)

        if pretty_res.succeeded:
            pretty_last = pretty_res.split(chr(61))[-1]
            """ Both types of quotes stripped below """
            pretty_name = pretty_last.strip('"\\\'')

            """ /etc/os-release found and examined """
            pname_array = pretty_name.split()
            pretty_name_id = pname_array[0].lower()
            if 'GNU' in pretty_name:
                """ Last word """
                pretty_name_word = pname_array[-1].strip('()')
            else:
                """ 2nd word / middle word """
                pretty_name_word = pname_array[1]
        else:
            pass

    elif pretty_res.succeeded:
        """ /etc/system-release existed so we are Rpm based probably """
        pname_array = pretty_name.split()
        pretty_name_id = pname_array[0].lower()
        version_major = 0
        for item in pname_array:
            try:
                version = float(item)
                version_major = int(version)
                break
            except ValueError, TypeError:
                pass
        if version_major > 0:
            pretty_name_word = 'unindexed'
            if version_major >= 0 and version_major < 21:
                pretty_name_word = NUM_AS_WORDS_TO_TWENTY[version_major]
    else:
        pass

    """ Example1: ('Debian GNU/Linux 7 (wheezy)', 'debian', 'wheezy')
    Example2: ('CentOS release 6.5 (Final)', 'centos', 'six')
    Example3: ('Ubuntu precise (12.04.3 LTS)', 'ubuntu', 'precise')
    """
    return (pretty_name, pretty_name_id, pretty_name_word)


def deb_or_rpm():
    package_type = None
    pretty_tuple = pretty_name_tuple()
    if pretty_tuple[1] == 'debian' or pretty_tuple[1] in DEBIAN_PLUS_DERIVATIVES2:
        package_type = 'deb'
    elif pretty_tuple[1] == 'centos' or pretty_tuple[1] in RPM_BASED2:
        package_type = 'rpm'
    else:
        package_type = None
    return package_type


def free_marked(print_flag=True):
    swap_total = None
    with hide('output'):
        pretty_res = run(CMD_FREE_M_T)
    free_lines = run_res_lines_stripped(pretty_res)
    print_lines = []
    if len(free_lines) > 0:
        for line in free_lines:
            if line.lower().startswith('swap'):
                swap_figures = line.split(chr(58))[1]
                swap_figures_array = swap_figures.strip().split()
                swap_total = swap_figures_array[0].strip()
                if swap_total == '0':
                    print_lines.append("{0}   ZEROSWAP".format(line))
                else:
                    print_lines.append("{0}   SWAPAVAILABLE".format(line))
            elif line.lower().startswith('total') and chr(58) in line:
                proc_res = run(CMD_NPROC)
                proc_count = 0
                if proc_res.succeeded:
                    try:
                        proc_count = int(str(proc_res).strip())
                    except Exception:
                        proc_count = 0
                if proc_count > 0:
                    print_lines.append("{0}   PROCESSORS:{1}".format(line,proc_count))
                else:
                    print_lines.append(line)
            else:
                print_lines.append(line)
    if print_flag:
        for line in print_lines:
            print line
    return swap_total


def md5_or_sudo(filepath):
    md5line = "{0} {1};".format(CMD_MD5SUM,filepath)
    md5res = run_or_sudo(md5line)
    md5lines = run_res_lines_stripped(md5res)
    mdf = None
    for line in md5lines:
        line_array = line.split()
        if len(line_array) > 1 and line_array[1] == filepath:
            mdf = line_array[0]
    return mdf


def md5text_or_sudo(filepath):
    md5line = "{0} {1};".format(CMD_MD5SUM_TEXT,filepath)
    md5res = run_or_sudo(md5line)
    md5lines = run_res_lines_stripped(md5res)
    mdf = None
    for line in md5lines:
        line_array = line.split()
        if len(line_array) > 1 and line_array[1] == filepath:
            mdf = line_array[0]
    return mdf


def put_using_named_temp_sudo(remote_filepath,lines):
    """ Variant of put_using_named_temp() which:
    (1) returns False on fail rather than using an abort and cleanup
    (2) Always uses sudo
    (3) Has two arguments rather than four arguments
    (4) Hands off to .write() appropriately for a string or iterable tuple/list
    (5) Lives here rather than files_directories so no circular dependency
    """
    """ remote_filepath should be a full path
    Example: '/etc/apache2/ports.conf.append'
    See the fabric put api for clarification.
    """
    import tempfile
    put_res = None
    put_return = True
    named_temp = tempfile.NamedTemporaryFile()
    try:
        if hasattr(lines, "__iter__"):
            for line in lines:
                named_temp.write(line)
        else:
            named_temp.write(lines)
        named_temp.flush()
        put_res = put(named_temp,remote_filepath,use_sudo=True)
    except Exception as e:
        raise
        put_return = False
    finally:
        named_temp.close()

    if put_res:
        if put_res.failed:
            put_return = False
        else:
            put_return = True
            """ Above else: for clarity rather than necessity """
    else:
        put_return = False
    return put_return

