#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2013 Gary Wright wrightsolutions.co.uk/contact
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# Tested against fabric 1.4.3
# http://ftp.uk.debian.org/debian/pool/main/f/fabric/fabric_1.4.3-1.debian.tar.gz

from __future__ import with_statement
from fabric.api import abort, cd, env, get, local, put, run, settings, sudo
from fabric.contrib.console import confirm
#from fabric.contrib.files import exists,append
from os import path,unlink
import re
import tempfile
try:
    from . import cmdbin
except ValueError:
    import cmdbin

from collections import OrderedDict
from mimetypes import add_type,guess_type



class Keyedfile(object):
    """ Abstraction of a simple text file format of key value pairs.
    Typical example (2 line file): VALUE1: on
    VALUE2: off
    Looser example (1 line file): VALUE1 on
    This is not intended to replace pre templated systems such as 'augeas'
    and for Apache2 configurations and sshd (pre templated) you might be better
    instead depending on augeas.

    Implementation: Intended to be used in a context ('with') and uses a class
    Keyvalfile() to hold keys and values.
    Keyvalfile() extends OrderedDict and will typically not have as many keys
    as there are lines in the file.
    However Keyvalfile() internally holds additional list and tuple helper variables
    which can be used to reconstruct the original file if you so wish (See limitations below)

    Configuration files on GNU/Linux and some unix/derivatives do cope with duplicate key
    values by allowing later values to supersede [redefine] a directive [key].
    With this in mind it would be preferable if any update method [sed] operated on
    the last line of the file with a matching key.

    Full implementation would support updates in addition to queries, but initially
    just the framework and queries are implemented.

    Limitations and recommended usage:
    The internal data structures are intended to:
    (.) support queries
    (.) support a reliable write of the file including any updates
    The data structures to support both are linked but separate.
    Do the queries first. Do use the reliable write features (when implemented)
    However you are discouraged from performing updates and then attempt to query things further.

    No contract is given here that detailed queries will be reliable once updates
    have taken place. Reiterating to clarify: Query, Update (optional), Write the changes

    By writing this function in a way that encourages context ('with') usage, there
    should be no encouragement to keep this object hanging around once your queries are
    complete, or you have updated then written.

    Limitation around toggling and case:
    on can be toggled to off and vice versa, however we currently make no provision
    for On to be toggled to Off where mixed case is used in the original input file.
    If this were to be added then certainly the function false_true_index() would
    require some thought. Would you accept On being toggle to all lowercase 'off'?

    Two update methods are planned 'put' and 'sed'. 
    In terms of number of interactions with the remote system,
    'put' is more lightweight than 'sed'.
    'sed' involves pushing a sed file to the remote and making the changes at the remote end.

    Design decision: The 'sed' method will now not automatically apply the updates on __exit__
    from the context, and will instead expect the user to be involved in a two stage process,
    the second of which is sending those sed lines to function sed_remote_apply()

    You might prefer the 'sed' method if you see the benefit of the additional code / checks
    in sed_remote_apply(), but do choose the 'put' method if you prefer a lighter touch.
    """

    MAX_LINELENGTH = 500
    MAX_LINES = 1000
    """ Above class variables are defined once for the class, but in practice
    accessed via self.MAX...
    The intention is to have a hard limit across the class, to help prevent
    accidental reading of things that might lead to memory exhaustion.
    But if you override or redefine things at runtime, then that is at your risk.
    """
    KEYSEP_NAMED = None
    """ One of several methods of removing the guesswork about what separates a
    key from a value, is to subclass and override KEYSEP_NAMED giving it the
    key separator you wish to force use of during parsing.
    """

    def __init__(self, filepath, markers=False, metrics=False, change_method='sed'):
        self._keyvalfile = None
        self._filepath = filepath
        self._markers = markers
        self._metrics = metrics
        self._change_method = change_method
        self._bufio = None
        self._given_linecount = 0

        given_lines = []
        if self.filepath is not None and hasattr(self.filepath, "__iter__"):
            """ Test / debug invocation passing in a list of lines rather than a filename """
            for line in self.filepath:
                self._given_linecount = 1 + self._given_linecount
                if line.endswith(chr(10)):
                    given_lines.append(line)
                else:
                    """ Add a line feed manually so truly simulates the
                    reading from file code involving StringIO """
                    given_lines.append("{0}{1}".format(line,chr(10)))
        elif self.filepath is not None:
            try:
                from StringIO import StringIO
                buf = StringIO()
                get_res = get(self.filepath,buf)
                buf.seek(0)
                line = True
                while line:
                    line = buf.readline()
                    self._given_linecount = 1 + self._given_linecount
                    given_lines.append(line)
                self._bufio = buf
            except IOError:
                raise Exception('IOError encountered - does file exist?')
        else:
            given_lines = []

        self._given_lines = given_lines
        given_lines = None

        if self._bufio is not None:
            """ We are not Test / debug invocation so do some extra checking on sanity of file """

            for ext in ['custom','ini','local','shared']:
                add_type('text/plain',".{0}".format(ext))

            guessed = guess_type(self._filepath)
            #print guessed

            filepath_dir = path.dirname(self._filepath)
            filepath_base = path.basename(self._filepath)

            if guessed[0] is None:
                """ If in common configuration directories /etc/default or /etc/sysconfig
                then assume okay for now
                """
                if filepath_dir == '/etc/default/':
                    pass
                elif filepath_base == 'resolv.conf':
                    pass
                elif self._filepath.startswith('/etc/sysconfig/'):
                    pass
                elif self._filepath.startswith('/etc/resolvconf/'):
                    pass
                elif filepath_base.endswith('_config'):
                    pass
                else:
                    """ Check nonbinary and for equals (=) or colon (:) """
                    grep_line = "{0} '{1}|{2}' {3};".format(cmdbin.CMD_EGREPI,chr(61),
                                                            chr(58),self._filepath)
                    if cmdbin.run_or_sudo(grep_line,quiet=True):
                        pass
                    else:
                        raise Exception("Binary or undetermined: {0}".format(self._filepath))
                """
                try:
                    #get_res = get(self.filepath,buf)
                    #buf.seek(0)
                    #buf.readline()
                except Exception:
                    raise Exception("Binary or undetermined: {0}".format(self._filepath))
                """
            elif guessed[0].startswith('text/'):
                pass
            else:
                pass

        return


    """ Class method to handle some logic about circumstances where
    (key,value) should not survive to make up the underlying OrderedDict
    For specialist needs subclass and override this class method.
    """
    @classmethod
    def list_of_tuples_appended(cls,key,value,list_before_append):
        if key is None or len(key) < 1:
            """ Refuse to append None or blank keys """
            return list_before_append
        elif key == value:
            """ Default behaviour is not to append where key
            and value are exactly equal """
            return list_before_append
        else:
            pass
        list_returned = list_before_append[:]
        list_returned.append((key,value))
        return list_returned


    """ Class method to instruct on how to parse a comment line.
    Stored within the class so it stays with the class definition.
    Override this in a subclass if you have specialized parsing needs.
    """
    @classmethod
    def process_comment_line(cls,stripped,keysep):
        # chr(58) is colon (:)            ; chr(61) is equals (=)           ; chr(32) is space
        stripped_array = stripped.split()
        key = ''.join(stripped_array[:2]).lstrip()
        if keysep is None:
            value = ''.join(stripped_array[2:])
        else:
            if keysep in [chr(61),chr(58)] and keysep in key:
                """ resplit but this time use the keysep we were given """
                stripped_resplit = stripped.split(keysep)
                key = stripped_resplit[0]
                if len(key.split()[0]) == 1:
                    """ Start of line has space after first character. """
                    key_array = key.split()
                    """ For lines like   # orange:on   we should make
                    the key '#orange' rather than '# orange' """
                    if len(key_array) > 1:
                        key = "{0}{1}".format(key_array[0],''.join(key_array[1:]))
                    value = ''.join(stripped_resplit[1:]).strip()
                else:
                    value = ''.join(stripped_array[2:])
            else:
                value = ''.join(stripped_array[2:])
        return (key,value)


    """ Class method to instruct on how to parse a line (keysep is not known).
    Stored within the class so it stays with the class definition.
    Override this in a subclass if you have specialized parsing needs.
    """
    @classmethod
    def process_promising_line(cls,stripped):
        """ There is no keysep argument here as if keysep is known then you
        should probably instead be calling process_line().
        """
        # chr(58) is colon (:)            ; chr(61) is equals (=)           ; chr(32) is space
        keysep = None
        keyd = None
        stripped_array = stripped.split()
        stripped_array_len = len(stripped_array)
        if stripped_array_len == 3:
            """ Example1: ['pgsql.max_persistent', '=', '-1']
            Example2: ['AcceptEnv', 'LANG', 'LC_*']
            print stripped_array
            """
                        #if 'persistent' in line:
                        #    print stripped_array_len,stripped_array_len,stripped_array_len
                        #    print line
            """ Check for equals (=) first then colon (:) """
            if len(stripped_array[1]) == 1 and stripped_array[1] in ([chr(61),chr(58)]):
                keysep = stripped_array[1]
                """ resplit but this time use the keysep we now have """
                stripped_array = stripped.split(keysep)
                keyfull = stripped_array[0]
                """ keyd (next) is stripped version """
                keyd = keyfull.strip()
                value = ''.join(stripped_array[1:]).strip()
            else:
                keyfull = stripped_array[0]
                value = ''.join(stripped_array[1:]).strip()
                """ For this sort of line no need to call tuple7, just construct manually """
                #list_of_tuples_all.append((keyfull,line,0,None,0,line[-1],linecount))
        elif stripped_array_len > 3:
            """ Example: ['#Privilege', 'Separation', 'is', 'turned', 'on', 'for', 'security']
            print stripped_array
            """
            keyfull = stripped_array[0]
            value = ''.join(stripped_array[1:]).strip()
            """ For this sort of line no need to call tuple7, just construct manually """
            #list_of_tuples_all.append((keyfull,line,0,None,0,line[-1],linecount))
        else:
            """ Probably of form key=value or key:value with no spacing around =/:
            Check for equals (=) first then colon (:) then finally space """
            if chr(61) in stripped_array[0]:
                keysep = chr(61)
                keyplus = stripped_array[0]
                pos = keyplus.index(keysep)
                keyfull = keyplus[:pos]
                """ keyd (next) is stripped version """
                keyd = keyfull.strip()
                stripped_array = stripped.split(keysep)
                value = ''.join(stripped_array[1:]).strip()
            elif chr(58) in stripped_array[0]:
                keysep = chr(58)
                keyplus = stripped_array[0]
                pos = keyplus.index(keysep)
                keyfull = keyplus[:pos]
                """ keyd (next) is stripped version """
                keyd = keyfull.strip()
                #value = keyplus[pos+1:]
                stripped_array = stripped.split(keysep)
                value = ''.join(stripped_array[1:]).strip()
            elif stripped_array_len == 2:
                keysep = chr(32)
                keyfull = stripped_array[0]
                value = ''.join(stripped_array[1:]).strip()
            else:
                keyfull = stripped_array[0]
                value = ''.join(stripped_array[1:]).strip()
                if keysep is not None:
                    """ resplit but this time use the keysep we now have """
                    stripped_array = stripped.split(keysep)
                    keyfull = stripped_array[0]
                    value = ''.join(stripped_array[1:]).strip()
                    """ No need to call tuple7, just construct manually """
                    #list_of_tuples_all.append((keyfull,line,0,keysep,0,line[-1],linecount))
        if keyd is None and keyfull is not None:
            """ If you have not assigned keyd explicitly above
            then it is simply same value as keyfull """
            keyd = keyfull
        return (keyfull,value,keysep,keyd)


    """ Class method to instruct on how to parse a line (keysep IS KNOWN).
    Stored within the class so it stays with the class definition.
    Override this in a subclass if you have specialized parsing needs.
    """
    @classmethod
    def process_line(cls,stripped,keysep):
        """ keysep argument is mandatory here. If keysep is not known
        should probably instead be calling process_promising_line().
        """
        # chr(58) is colon (:)            ; chr(61) is equals (=)           ; chr(32) is space
        if keysep is None:
            raise Exception('Parse: Only process_promising_line() should be called keysep=None')
        stripped_array = stripped.split(keysep)
        keyfull = stripped_array[0]
        """ keyd (next) is stripped version """
        keyd = keyfull.strip()
        value = ''.join(stripped_array[1:]).strip()
        #if keyd == 'orange':
        #    print "{0}F{1}D{2}".format(keyfull,keyd,value)
        """ The spacing preserved in keyfull will be used later within tuple7() 
        to record before_count and after_count so that we can define very
        focussed sed targets for any requested toggle / replacements """
        return (keyfull,value,keysep,keyd)


    """ Following properties have no set() as only set directly during class init. """
    @property
    def keyvalfile(self):
        return self._keyvalfile

    @property
    def filepath(self):
        return self._filepath

    @property
    def markers(self):
        return self._markers

    @property
    def metrics(self):
        return self._metrics

    @property
    def change_method(self):
        return self._change_method


    def tuple7(self,key_unstripped,line,linenum,keysep=None):
        """ Returns a 7 tuple with fields as follows:
        key, line, before_count, sep, after_count, linesep, linenum

        The first argument should include any right padding of spaces,
        as we will be recording the count of that padding in before_count
        although we will strip the key before then placing it in the tuple.
        """
        appended = True
        unstripped_len = len(key_unstripped)
        key_rstripped = key_unstripped.rstrip()
        stripped_len = len(key_rstripped)
        before_count = 0
        if unstripped_len > stripped_len:
            before_count = unstripped_len - stripped_len
        after_count = 0
        """ after_count = 0 (above) as would always be zero when keysep is space """
        if keysep is not None and len(keysep.strip()) > 0:
            remainder = line[(unstripped_len+len(keysep)):]
            remainder_len = len(remainder)
            stripped_len = len(remainder.lstrip())
            if remainder_len > stripped_len:
                after_count = remainder_len - stripped_len
        return (key_rstripped,line,before_count,keysep,after_count,line[-1],linenum)


    def __enter__(self):


        class Keyvalfile(OrderedDict):

            FALSELIST = ['0','false','False','off','no']
            TRUELIST = ['1','true','True','on','yes']

            def __init__(self, *args, **kwargs):
                super(Keyvalfile, self).__init__(*args, **kwargs)
                self._filepath = None
                self._linecount = 0
                self._linecount_adjustment = 0
                self._lines_tuple = ()
                """ Above is a tuple of lines, below is a list of 7 field tuples """
                self._tuples_all = []
                self._conf_linesep = None
                self._change_count = 0
                self._md5sum = 'a'*32
                self._keysep = ' '
                self._dict_of_changed = OrderedDict()
                """ Regardless of update method, dict_of_changed and dict_of_seds
                are kept in step remember. """
                self._dict_of_seds = OrderedDict()
                self._keylen_average = 0
                """ Duplicates will not play any part in boosting or reducing the figure
                for keylen_average as we are sourcing direct from the dictionary object
                itself.
                Two reasons why keylen_average_full might be higher than keylen_average:
                (.) keylen_average_full takes into account every non blank key and
                will not exclude duplicate keys from the totalling process.
                (.) Some implementations of the parser routines may choose to strip
                (leading strip?) the key before adding it to the dictionary.
                keylen_average_full in that case may be counting spacing which
                will not be counted when totalling directly from the dictionary keys
                """
                self._keylen_average_full = 0
                """ See the traffic light examples in the test suite for an example
                of input that would produce key_duplicates_count > 0 """
                self._key_duplicates_count = 0
                """ keylist_frozen is a one time snapshot of the keys when the object
                was populated. It will never be updated. Use of self.items() is preferable
                unless you have a discrepancy and feel consulting the snapshot would help
                """
                self._keylist_frozen = []


            """ linecount property requires no setter as only when lines supplied """
            @property
            def linecount(self):
                return self._linecount

            def linecount_adjustment_get(self):
                return self._linecount_adjustment

            #def linecount_adjustment_set(self,linecount_adjustment):
            #    self._linecount_adjustment = linecount_adjustment

            def linecount_adjustment_decrement(self):
                self._linecount_adjustment = self.linecount_adjustment_get()-1

            def linecount_adjustment_increment(self):
                self._linecount_adjustment = 1+self.linecount_adjustment_get()

            linecount_adjustment = property(linecount_adjustment_get)

            def linecount_adjusted():
                return (self._linecount + self._linecount_adjustment)

            @property
            def lines_tuple(self):
                """ Tuple of lines (never updated). See tuples_all if you want a list of 7 field tuples
                Call lines_for_put() if you want something might include changes that you initiated. """
                return self._lines_tuple

            @property
            def tuples_all(self):
                """ tuples_all contains tuples of 7 fields:
                key, line, before_count, sep, after_count, linesep, linenum
                See lines_tuple if you instead want a tuple of lines """
                return self._tuples_all

            @property
            def change_count(self):
                return self._change_count

            @property
            def conf_linesep(self):
                return self._conf_linesep

            def change_count_increment(self):
                self._change_count += 1

            def switch_commented_on_toggle_get(self):
                return self._switch_commented_on_toggle

            def switch_commented_on_toggle_set(self,switch_commented_on_toggle):
                self._switch_commented_on_toggle = switch_commented_on_toggle

            switch_commented_on_toggle = property(switch_commented_on_toggle_get,switch_commented_on_toggle_set)


            """ md5sum property requires no setter as only set when lines supplied """
            @property
            def md5sum(self):
                return self._md5sum

            """ keysep property requires no setter as only set when lines supplied """
            @property
            def keysep(self):
                return self._keysep

            """ dict_of_seds property requires no setter as only set when lines supplied """
            @property
            def dict_of_seds(self):
                """ Dictionary of seds. Key is line number of original file line on which the
                sed is intended to operate. Value is the sedline to feed to the sed utility
                """
                return self._dict_of_seds

            """ keylen_average property requires no setter as only set when lines supplied """
            @property
            def keylen_average(self):
                return self._keylen_average

            """ keylen_average_full property requires no setter as only set when lines supplied """
            @property
            def keylen_average_full(self):
                return self._keylen_average_full

            """ key_duplicates_count property requires no setter as only set when lines supplied """
            @property
            def key_duplicates_count(self):
                return self._key_duplicates_count

            """ keylist_frozen property requires no setter as only set when lines supplied """
            @property
            def keylist_frozen(self):
                return self._keylist_frozen


            """ list_of_sedlines property requires no setter as only constructed at call time """
            @property
            def list_of_sedlines(self):
                """ If we had something that looked like this: 2 s/bb/tt/; 1 s/aa/ss/; 3 s/cc/uu/
                Then we used the linenumber (in this case 2,1,3) and order things in descending order
                so that our resulting list is like: s/cc/uu/; s/bb/tt/ ; s/aa/ss/
                """
                sedlines = []
                for linenum in sorted(self._dict_of_seds.iterkeys(),reverse=True):
                    sedlines.append(self._dict_of_seds[linenum])
                return sedlines


            def enclosed_by_char(self,val):
                enclosed_by = None
                given_len = len(val)
                val_stripped = val.strip(chr(34))
                if 2 == given_len - len(val_stripped):
                    """ enclosed by double quotes """
                    enclosed_by = chr(34)
                else:
                    val_stripped = val.strip(chr(39))
                    if 2 == given_len - len(val_stripped):
                        """ enclosed by single quotes """
                        enclosed_by = chr(39)
                return enclosed_by


            def tuple_first_field_equals(self,tup,target):
                equality = False
                first = tup[0]
                #print first,target
                if first == target:
                    equality = True
                return equality


            def tuples_all_from_key(self,target_key,reverse=True):
                """ reversed=True is the appropriate defaut to fit with the
                contract of these classes. Extract from earlier note at top of outer class Keyedfile():
                ...by allowing later values to supersede [redefine] a directive [key] ...
                This function may well be called more often than the similar function tuple_all_line_from_key()
                Employs an alternative approach: Use ifilter rather than the reversed(filter( 
                code used in tuple_all_line_from_key()
                tuples_all contains tuples of 7 fields:
                key, line, before_count, sep, after_count, linesep, linenum
                """
                from itertools import ifilter
                from functools import partial
                tup = None
                try:
                    tuples = ifilter(partial(self.tuple_first_field_equals,target=target_key),
                                     self.tuples_all)
                except TypeError:
                    tuples = None
                    raise
                try:
                    tuples_list = list(tuples)
                    """ Assign above to a variable so no problem consuming twice """
                    tup = tuples_list[0]
                    if reverse:
                        tup = tuples_list[-1]
                except Exception:
                    tup = None
                return tup


            def tuples_all_line_from_key(self,target_key,reverse=True):
                """ reversed=True is the appropriate defaut to fit with the
                contract of these classes. Extract from earlier note at top of outer class Keyedfile():
                ...by allowing later values to supersede [redefine] a directive [key] ...
                """
                line = None
                if reverse:
                    tuples_all_not_none = reversed(filter(None,self.tuples_all))
                else:
                    tuples_all_not_none = filter(None,self.tuples_all)
                for tup in tuples_all_not_none:
                    if tup[0] == target_key:
                        line = tup[1]
                        break
                return line


            def sedqueue(self,linenum,sedline):
                queue_return = True
                if sedline is None or len(sedline) < 1 or sedline.startswith('q'):
                    queue_return = False
                elif linenum in self._dict_of_seds:
                    """ False if we already have a sed operation stored for this file line """
                    queue_return = False
                elif sedline in self.list_of_sedlines:
                    """ False if we already have this sedline stored """
                    queue_return = False
                else:
                    """ Do a simple check to make sure the dictionaries are going to
                    be in step, and refuse to queue if there is an issue
                    """
                    #print sedline
                    dict_diff = (len(self._dict_of_changed)-len(self._dict_of_seds))
                    if dict_diff == 1:
                        self._dict_of_seds[linenum]=sedline
                        self._change_count = self._change_count + 1
                    else:
                        queue_return = False
                return queue_return


            def dequeue_all(self):
                dequeued_count = 0
                sedline_count = len(self._dict_of_seds)
                if sedline_count > 0:
                    self._dict_of_changed = OrderedDict()
                    self._dict_of_seds = OrderedDict()
                    dequeued_count = sedline_count
                    self._change_count = 0
                return dequeued_count
                  

            def linecount_and_lines_set(self,count,lines,tuples_all,keysep=None):
                set_flag = False
                if len(lines) == count:
                    self._linecount = count
                    self._lines_tuple = tuple(lines)
                    self._tuples_all = tuples_all
                    """ lines_tuple is a fixed record of the original input file.
                    tuples_all is also a fixed record but in the form of tuples
                    where the first field represents a recognisable key (if any)
                    and the second field is the original input line.
                    The tuples in tuples_all have several fields to aid later processing:
                    key, line, before_count, sep, after_count, linesep, linenum
                    """
                    if keysep is not None:
                        self._keysep = keysep
                        key_duplicates_count = 0
                        keylist_frozen = []
                        """ Next we attempt to measure average key length, but
                        only include in our reckoning those lines which were
                        marked as having a key separator (stored in 4th field)
                        """
                        lines_counted = 0
                        lines_total = 0
                        # tuple: key, line, before_count, sep, after_count, linesep, linenum
                        for tup in tuples_all:
                            key = tup[0]
                            if key is not None:
                                if key in keylist_frozen:
                                    key_duplicates_count += 1
                                else:
                                    keylist_frozen.append(key)
                                if tup[3] is not None:
                                    # Maybe wise to avoid: len(key.strip()) for key_average_full
                                    lines_total += len(key)
                                    lines_counted += 1
                        if lines_counted > 0:
                            self._keylen_average_full = int(round(lines_total/float(lines_counted)))
                        """ keylen_average_full now populated so turn our attention to keylen_average
                        which is the average of the keys as they are stored in the OrderedDict.
                        Where we have chosen to apply additional stripping before storing the key
                        in OrderedDict, we may see a lower average figure next """
                        conf_len = len(self)
                        if conf_len > 0:
                            keys_total = 0
                            for dict_key in self:
                                if dict_key is not None:
                                    keys_total += len(dict_key)
                            """ Duplicates will not play any part in boosting or reducing the figure
                            for keylen_average as we are sourcing direct from the dictionary object
                            itself.
                            Two reasons why keylen_average_full might be higher than keylen_average:
                            (.) keylen_average_full takes into account every non blank key and
                            will not exclude duplicate keys from the totalling process.
                            (.) Some implementations of the parser routines may choose to strip
                            (leading strip?) the key before adding it to the dictionary.
                            keylen_average_full in that case may be counting spacing which
                            will not be counted when totalling directly from the dictionary keys
                            """
                            self._keylen_average = int(round(keys_total/float(conf_len)))

                    self._key_duplicates_count = key_duplicates_count
                    self._keylist_frozen = keylist_frozen[:]
                    set_flag = True
                    self._lines = lines[:]
                    from hashlib import md5
                    mdf = md5()
                    last_dict = {}
                    for line in lines:
                        mdf.update(line)
                        if len(line) > 0:
                            last = line[-1]
                            if last in last_dict:
                                last_dict[last]=1+last_dict[last]
                            else:
                                last_dict[last]=1
                    self._md5sum = mdf.hexdigest()
                    self._conf_linesep = max(last_dict,key=last_dict.get)
                return set_flag
                

            def keyval(self,key):
                """ User should probably understand that we are OrderedDict but 
                provide a convenient helper also """
                return self.get(key)


            def keyline(self,key):
                return self.get(key)


            def false_true_index(self,val_given):
                """ If val is in a predefined FALSE list, then return a three tuple of:
                (False,index,enclosed_by) where enclosed_by is None if no enclosing quotes
                If val is in a predefined TRUE list, then return a three tuple of:
                (True,index,enclosed_by) where enclosed_by is None if no enclosing quotes
                If we are in neither list then return (None,-1,None)
                """
                return_tuple = (None,-1,None)
                if val_given in Keyvalfile.FALSELIST:
                    idx = Keyvalfile.FALSELIST.index(val_given)
                    return_tuple = (False,idx,None)
                elif val_given in Keyvalfile.TRUELIST:
                    idx = Keyvalfile.TRUELIST.index(val_given)
                    return_tuple = (True,idx,None)
                else:
                    enclosed_by = self.enclosed_by_char(val_given)
                    if enclosed_by is not None:
                        val_stripped = val_given.strip(enclosed_by)
                        if val_stripped in Keyvalfile.FALSELIST:
                            idx = Keyvalfile.FALSELIST.index(val_stripped)
                            return_tuple = (False,idx,enclosed_by)
                        elif val_stripped in Keyvalfile.TRUELIST:
                            idx = Keyvalfile.TRUELIST.index(val_stripped)
                            return_tuple = (True,idx,enclosed_by)
                        else:
                            pass

                return return_tuple


            def record_changed(self,target_key,target,replacement,lineend_check=True):
                recorded = False
                if len(target) < len(target_key):
                    return recorded
                tup7 = self.tuples_all_from_key(target_key)
                assert len(tup7) > 6
                if tup7 is None:
                    return recorded
                linenum = tup7[6]
                line = tup7[1]
                matched = re.match(target,line)
                if matched:
                    if lineend_check:
                        """ When lineend_check is set True, do your best to ensure any
                        line ending in the original line is in replacement also. """
                        try:
                            if replacement[-1] == line[-1]:
                                pass
                            else:
                                replacement = "{0}{1}".format(replacement,line[-1])
                        except Exception:
                            pass
                    line_changed = "{0}{1}".format(line[:matched.end()],replacement)
                    self._dict_of_changed[linenum]=line_changed
                    recorded = True
                return recorded


            def target_spaced(self,target_key,hatted=True):
                """ Examines the spacing of the existing line indicated by target_key
                and produces a sed target string that will faithfully represent the
                current target including appropriate keysep and spacing.
                Except where hatted=False assume that we should prefix the sed target with ^
                """
                spaced = target_key
                if len(self.keysep.strip()) < 1:
                    pass
                else:
                    spaced = "{0}{1}".format(target_key,self.keysep)
                    """ Probably equals (=) or colon (:) used to separate keys from values.
                    Begin with a lookup of the original line by consulting self.tuples_all
                    tuples_all contains tuples of 7 fields:
                    key, line, before_count, sep, after_count, linesep, linenum
                    """
                    tup = self.tuples_all_from_key(target_key)
                    if tup is not None:
                        #print tup
                        if tup[0] is None:
                            raise Exception('Supplied target_key lookup did not find expected tuple')
                        elif tup[3] == self.keysep:
                            line = tup[1]
                            pos = line.index(self.keysep)
                            after_count = tup[4]
                            if after_count > 0:
                                spaced = "{0}{1}".format(line[:(pos+len(self.keysep))],' '*after_count)
                            else:
                                spaced = line[:(pos+len(self.keysep))]
                            line_array = line.split(self.keysep)
                        else:
                            raise Exception("Keyed file line uses {0} instead of expected sep={1}".format(
                                    tup[3],self.keysep))
                if hatted:
                    spaced = "^{0}".format(spaced)
                return spaced


            def sedline(self,target_key,required_val):
                """ Lightweight version of sedline_safe which is really for previewing
                changes only, as it does not populate dict_of_changed().
                For files with key and value separated by space (so
                self.kepsep is not (=) or (:) and you do not want safety of specifying
                required_val, then this is a lightweight alternative that should do just fine.
                Also use sedline_safe() if you your file has values wrapped in double quotes or single
                quotes rather than being bare. Example ENABLED="false" rather than ENABLED false
                """
                constructed = ''
                if target_key is not None and len(target_key) > 1:
                    """ Use \\1 below when really we mean \1 so as Python treats properly """
                    if len(self.keysep.strip()) < 1:
                        constructed = "0,/{0}\s+/ s/\({0}\)\(.*\)/\\1 {1}/".format(
                            target_key,required_val)
                    else:
                        """ keysep is equals or colon or something other than space(s) """
                        constructed = "0,/{0}{1}/ s/\({0}{1}\)\(.*\)/\\1{2}/".format(
                            target_key,self.keysep,required_val)
                return constructed


            def sedline1(self,target_key,required_val,enclosed_by=None):
                """ Variant of sedline1 for previewing / test results in a fashion
                that is a little closer to sedline_safe() but without updating dict_of_changed().
                """
                constructed = ''
                if target_key is not None and len(target_key) > 1:
                    if enclosed_by is None:
                        """ Use \\1 below when really we mean \1 so as Python treats properly """
                        if len(self.keysep.strip()) < 1:
                            constructed = "0,/{0}\s+/ s/\({0}\)\(.*\)/\\1 {1}/".format(
                                target_key,required_val)
                        else:
                            """ keysep is equals or colon or something other than space(s) """
                            constructed = "0,/{0}/ s/\({0}\)\(.*\)/\\1{1}/".format(
                                self.target_spaced(target_key),required_val)
                    else:
                        stored_val = self[target_key]
                        #if stored_val.startswith(enclosed_by):
                        """ Use \\1 below when really we mean \1 so as Python treats properly """
                        if len(self.keysep.strip()) < 1:
                            constructed = "0,/{0}\s+/ s/\({0}\)\(.*\)/\\1 {2}{1}{2}/".format(
                                target_key,required_val,enclosed_by)
                        else:
                            """ keysep is equals or colon or something other than space(s) """
                            constructed = "0,/{0}/ s/\({0}\)\(.*\)/\\1{2}{1}{2}/".format(
                                self.target_spaced(target_key),required_val,enclosed_by)
                return constructed


            def sedline_safe(self,target_key,required_val,enclosed_by=None):
                """ Use this function when you have a firm intention to update the
                file (rather than just previewing) as calling this function will
                record your intention in dict_of_changed().
                For simple previewing use lightweight function sedline() instead.
                If you wanted to make the new value (including single quotes)
                to be 'false' instead of existing value 'true', then use this function
                with final argument set to single quote and give the unquoted value
                as required_val parameter.
                The 'safe' in the function name refers to the startswith() check
                on the existing value otherwise sedline returned is empty.
                Unlike the lightweight sedline() function we use the function target_spaced()
                to produce a faithful sed target when keysep is other than space.
                """
                constructed = ''
                if target_key is None:
                    return constructed
                if len(target_key) < 2:
                    return constructed
                replacement = 'errorval'
                if enclosed_by is None:
                    replacement = required_val
                else:
                    stored_val = self[target_key]
                    if stored_val.startswith(enclosed_by):
                        replacement = "{0}{1}{0}".format(enclosed_by,required_val)
                    else:
                        replacement = 'errorunsafe'
                if replacement.startswith('error'):
                    return constructed
                if len(self.keysep.strip()) < 1:
                    replacement = " {0}".format(replacement)
                """ Preliminary logic complete, now form 'constructed' """
                target = target_key
                if len(self.keysep.strip()) > 0:
                    target = self.target_spaced(target_key)
                constructed = "0,/{0}/ s/\({0}\)\(.*\)/\\1{1}/".format(target,replacement)
                """ Above we use \\1 below when really we mean \1 so as Python expresses it properly """
                if self.record_changed(target_key,target,replacement) is False:
                    """ If you cannot make sense of the planned change or recording it fails then
                    abandon your construction and set constructed to empty """
                    constructed = ''
                return constructed


            def tuple_disabled(self,tup,return_type='flag'):
                """ Takes a two tuple (Example: an entry from tuples_all) and 
                based on the value of first field returns False or the second field
                When return_type is not 'value' then only possible returns are True or False """
                if tup[0] is None:
                    return False
                if not tup[0].startswith(chr(35)):
                    return False
                """ Have got this far so tup[0] must startwith '#' """
                if return_type == 'value':
                    return tup[1]
                return True


            def tuple_enabled(self,tup,return_type='flag'):
                """ Takes a tuple (Example: an entry from tuples_all) and 
                based on the value of first field returns False or the second field
                When return_type is not 'value' then only possible returns are True or False """
                if tup[0] is None:
                    return False
                if tup[0].startswith(chr(35)):
                    return False
                if return_type == 'value':
                    return tup[1]
                return True


            def tuples_enabled(self):
                """ Return a list of tuples. Formed from tuples_all where
                we have filtered and only retained tuples where key is enabled """
                tuples = []
                for tup in self.tuples_all:
                    if tup[0] is None or tup[0].startswith(chr(35)):
                        pass
                    else:
                        tuples.append(tup)
                return tuples


            def tuples_without_keysep(self,enabled=True):
                """ Tuple: key, line, before_count, sep, after_count, linesep, linenum """
                if enabled:
                    tuples = []
                    for tup in self.tuples_all:
                        if tup[0] is None or tup[0].startswith(chr(35)):
                            continue
                        if tup[3] is None:
                            tuples.append(tup)
                        else:
                            pass
                else:
                    tuples = []
                    for tup in self.tuples_all:
                        if tup[3] is None:
                            tuples.append(tup)
                        else:
                            pass
                return tuples


            def tuples_with_keysep(self,enabled=True,keysep='default'):
                """ Tuple: key, line, before_count, sep, after_count, linesep, linenum """
                if keysep is None:
                    return tuples_without_keysep()
                if keysep is 'default' and enabled:
                    tuples = []
                    for tup in self.tuples_all:
                        if tup[0] is None or tup[0].startswith(chr(35)):
                            pass
                        elif tup[3] is not None:
                            tuples.append(tup)
                        else:
                            pass
                elif enabled:
                    tuples = []
                    for tup in self.tuples_all:
                        if tup[0] is None or tup[0].startswith(chr(35)):
                            continue
                        tuple_sep = tup[3]
                        if tuple_sep == keysep:
                            tuples.append(tup)
                        else:
                            pass
                else:
                    tuples = []
                    for tup in self.tuples_all:
                        if tup[3] == keysep:
                            tuples.append(tup)
                        else:
                            pass
                return tuples


            def values_enabled(self,list_of_tups):
                """ Returns a list of values where the value is enabled.
                Takes a list of tuples as input. Output is a filtered list of values
                """
                list_returned = []
                for tup in list_of_tups:
                    if tup[0] is None:
                        continue
                    if tup[0].startswith(chr(35)):
                        continue
                    list_returned.append(tup[1])
                return list_returned


            def flags_enabled(self,list_of_tups):
                """ Returns a list of True / False indicating when the value is enabled.
                Takes a list of tuples as input. Output is an unfiltered list of True / False
                """
                list_returned = []
                for tup in list_of_tups:
                    if tup[0] is None:
                        list_returned.append(False)
                        continue
                    if tup[0].startswith(chr(35)):
                        list_returned.append(False)
                        continue
                    list_returned.append(True)
                return list_returned


            def keyvals_typed(self,linetype,keys_list):
                lines_returned = []
                if not hasattr(keys_list, "__iter__"):
                    print 'keyvals_typed() sees API misuse. An Iterable expected - not given'
                    return lines_returned
                if linetype is None:
                    linetype = 1
                else:
                    linetype = int(linetype)
                """ linetype=0 says return only keys in keys_list that are disabled.
                linetype=1 says return only keys in keys_list that are enabled.
                linetype > 1 says return keys in keys_list that are disabled or enabled.
                """
                if linetype < 1:
                    """ Only key lines where the key matches and disabled """
                    for target in keys_list:
                        key_target = "#{0}".format(target)
                        if key_target in self:
                            lines_returned.append(self.get(key_target))
                elif linetype == 1:
                    """ Only key lines where the key matches and enabled """
                    for key_target in keys_list:
                        if key_target in self:
                            lines_returned.append(self.get(key_target))
                else:
                    """ 'in self' as we are an OrderedList of course """
                    for target in keys_list:
                        if target in self:
                            lines_returned.append(self.get(target))
                        key_target = "#{0}".format(target)
                        if key_target in self:
                            lines_returned.append(self.get(key_target))
                return lines_returned


            def keylines_disabled(self):
                """ This function uses a fixed representation of the file
                and would NOT include any updates you have initiated.
                Returns a list of lines. Although sourced from tuples_all, we
                return just the original line and do nothing with the key """
                filtered = filter(self.tuple_disabled,self.tuples_all)
                """ filtered are two tuples so we will use a for loop below to 
                take just second field.
                Example of two tuple in filtered ('AcceptEnv', 'AcceptEnv LANG LC_*\n')
                """
                disabled = []
                for item in filtered:
                    disabled.append(item[1])
                """ Above append (for our example tuple) would append 'AcceptEnv LANG LC_*\n' """
                return disabled


            def keylines_enabled(self):
                """ This function uses a fixed representation of the file
                and would NOT include any updates you have initiated.
                Returns a list of lines. Although sourced from tuples_all, we
                return just the original line and do nothing with the key """
                filtered = filter(self.tuple_enabled,self.tuples_all)
                """ filtered are two tuples so we will use a for loop below to 
                take just second field.
                Example of two tuple in filtered ('AcceptEnv', 'AcceptEnv LANG LC_*\n')
                """
                enabled = []
                for item in filtered:
                    enabled.append(item[1])
                """ Above append (for our example tuple) would append 'AcceptEnv LANG LC_*\n'
                See lines_enabled() for a variant of this function that does rstrip()
                and will therefore remove the line ending character """
                return enabled


            def keylines_enabled_resplit(self):
                enabled = self.keylines_enabled()
                unsplit = ''.join(enabled)
                """ print "conf_linesep:{0} {0} {0}".format(self.conf_linesep) """
                return unsplit.split(self.conf_linesep)


            def keylines_typed(self,linetype,keys_list):
                lines_returned = []
                if not hasattr(keys_list, "__iter__"):
                    print 'keyvals_typed() sees API misuse. An Iterable expected - not given'
                    return lines_returned
                if linetype is None:
                    linetype = 1
                else:
                    linetype = int(linetype)
                """ linetype=0 says return only keys in keys_list that are disabled.
                linetype=1 says return only keys in keys_list that are enabled.
                linetype > 1 says return keys in keys_list that are disabled or enabled.
                """
                if linetype < 1:
                    """ Only key lines where the key matches and disabled """
                    for target in keys_list:
                        key_target = "#{0}".format(target)
                        if key_target in self:
                            for tup in self.tuples_all:
                                if tup[0] == key_target:
                                    lines_returned.append(tup[1])
                elif linetype == 1:
                    """ Only key lines where the key matches and enabled """
                    for key_target in keys_list:
                        if key_target in self:
                            for tup in self.tuples_all:
                                if tup[0] == key_target:
                                    lines_returned.append(tup[1])
                else:
                    for tup in self.tuples_all:
                        if tup[0] is not None:
                            if tup[0] in keys_list:
                                lines_returned.append(tup[1])
                            elif tup[0].lstrip(chr(35)) in keys_list:
                                lines_returned.append(tup[1])
                return lines_returned


            def keyvals(self,keys_list):
                return self.keyvals_typed(1,keys_list)


            def keylines(self,keys_list):
                return self.keylines_typed(1,keys_list)


            def lines_enabled(self):
                """ This function uses a fixed representation of the file
                and would NOT include any updates you have initiated.
                Returns a list of lines. Although sourced from tuples_all, we
                return just the original line and do nothing with the key """
                filtered = filter(self.tuple_enabled,self.tuples_all)
                """ filtered are two tuples so we will use a for loop below to 
                take just second field.
                Example of two tuple in filtered ('AcceptEnv', 'AcceptEnv LANG LC_*\n')
                """
                enabled = []
                for item in filtered:
                    enabled.append(item[1].rstrip())
                """ Above append (for our example tuple) would append 'AcceptEnv LANG LC_*'
                See keylines_enabled() for a variant of this function that does not rstrip()
                and will therefore retain the line ending character """
                return enabled


            def lines_enabled_contains(self,target):
                """ This function uses a fixed representation of the file
                and would NOT include any updates you have initiated.
                Returns a list of lines. Although sourced from tuples_all, we
                return just the original line and do nothing with the key.
                An example of where you might use this function is by setting
                target=' files' and examining /etc/nsswitch.conf
                if no target is given then the results would be as for lines_enabled()
                """
                if target == '':
                    """ Keeps the logic in the later loop straightforward """
                    target = None
                tuples = self.tuples_enabled()
                enabled = []
                for tup in tuples:
                    if target is None or target in tup[1]:
                        enabled.append(tup[1].rstrip())
                """ Above we have rstrip() the line to lose any line ending delimiter
                Use tuples_enabled() directly and loop over the results if you do
                not want the lines to have been right stripped. """
                return enabled


            def lines_current_simple(self):
                """ Unlike lines_for_put_simple, you will not find blank lines or
                comment lines in what is returned from this function. """
                if self.change_count < 1:
                    return self._lines_tuple
                """ Having got this far we need to overlay the original lines with any
                changes recorded in dict_of_changed """
                lines = []
                for tup7 in self._tuples_all:
                    linenum = tup7[6]
                    if linenum not in self._dict_of_changed:
                        continue
                    line_changed = self._dict_of_changed[linenum]
                    if line_changed is not None and len(line_changed.strip()) > 2:
                        lines.append(line_changed)
                    else:
                        lines = []
                        break
                return lines


            def lines_for_put_simple(self):
                """ Use self.lines_tuple if you want the original instead of current """
                if self.change_count < 1:
                    return self._lines_tuple
                """ Having got this far we need to overlay the original lines with any
                changes recorded in dict_of_changed """
                lines = []
                linenum = -1
                for line in self.lines_tuple:
                    linenum += 1
                    if linenum not in self._dict_of_changed:
                        lines.append(line)
                        continue
                    line_changed = self._dict_of_changed[linenum]
                    if line_changed is not None and len(line_changed.strip()) > 2:
                        lines.append(line_changed)
                        #print "Line {0} (changed) has been appended.".format(linenum)
                    else:
                        lines = []
                        break
                return lines


            def lines_for_put(self,simratio_quick=0,simratio=0):
                """ Use self.lines_tuple if you want the original instead of current """
                if self.change_count < 1:
                    return self._lines_tuple
                """ Having got this far we need to overlay the original lines with any
                changes recorded in dict_of_changed """
                lines = []
                if simratio_quick == 0 and simratio == 0:
                    return self.lines_for_put_simple()
                elif simratio_quick == 0:
                    from difflib import SequenceMatcher
                    for tup7 in self._tuples_all:
                        linenum = tup7[6]
                        if linenum not in self._dict_of_changed:
                            continue
                        line_changed = self._dict_of_changed[linenum]
                        if line_changed is not None and len(line_changed.strip()) > 2:
                            line_original = tup7[1]
                            sm = SequenceMatcher(None,line_original,line_changed)
                            if sm.ratio() >= simratio:
                                lines.append(line_changed)
                            else:
                                lines = []
                                break
                        else:
                            lines = []
                            break
                elif simratio == 0:
                    from difflib import SequenceMatcher
                    for tup7 in self._tuples_all:
                        linenum = tup7[6]
                        if linenum not in self._dict_of_changed:
                            continue
                        line_changed = self._dict_of_changed[linenum]
                        if line_changed is not None and len(line_changed.strip()) > 2:
                            line_original = tup7[1]
                            sm = SequenceMatcher(None,line_original,line_changed)
                            if sm.quick_ratio() >= simratio_quick:
                                lines.append(line_changed)
                            else:
                                lines = []
                                break
                        else:
                            lines = []
                            break
                else:
                    from difflib import SequenceMatcher
                    for tup7 in self._tuples_all:
                        linenum = tup7[6]
                        if linenum not in self._dict_of_changed:
                            continue
                        line_changed = self._dict_of_changed[linenum]
                        if line_changed is not None and len(line_changed.strip()) > 2:
                            line_original = tup7[1]
                            sm = SequenceMatcher(None,line_original,line_changed)
                            if sm.quick_ratio() >= simratio_quick and sm.ratio() >= simratio:
                                lines.append(line_changed)
                            else:
                                lines = []
                                break
                        else:
                            lines = []
                            break
                return lines


            def lines_for_put_auto(self):
                """ Use self.lines_tuple if you want the original instead of current """
                if self.change_count < 1:
                    return self._lines_tuple
                """ Having got this far we need to overlay the original lines with any
                changes recorded in dict_of_changed """
                RATIO_AUTO=0.5
                lines = []
                from difflib import SequenceMatcher
                for tup7 in self._tuples_all:
                    linenum = tup7[6]
                    if linenum not in self._dict_of_changed:
                        continue
                    line_changed = self._dict_of_changed[linenum]
                    if line_changed is not None and len(line_changed.strip()) > 2:
                        line_original = tup7[1]
                        sm = SequenceMatcher(None,line_original,line_changed)
                        if sm.ratio() >= RATIO_AUTO:
                            lines.append(line_changed)
                        else:
                            lines = []
                            break
                    else:
                        lines = []
                        break
                return lines


            def sedline_toggle(self,target_key):
                """ There is nothing to stop you using this function directly,
                however you are encouraged to use toggle_safe() wrapper as
                it can provide extra useful checks """
                sedline = None
                enclosed_by = None
                stored_val = self[target_key]
                tuple3 = self.false_true_index(stored_val)
                if tuple3[0] is not None:
                    if tuple3[0] is False:
                        target_val = Keyvalfile.TRUELIST[tuple3[1]]
                        sedline = self.sedline_safe(target_key,target_val,tuple3[2])
                    elif tuple3[0] is True:
                        target_val = Keyvalfile.FALSELIST[tuple3[1]]
                        sedline = self.sedline_safe(target_key,target_val,tuple3[2])
                return sedline

            
            def sedline_toggle_wrapper(self,target_key):
                """ There is nothing to stop you using this function directly,
                however you are encouraged to use toggle_safe() wrapper as
                it can provide extra useful checks """
                make_return = False
                if target_key in self:
                    sedline = self.sedline_toggle(target_key)
                    if sedline is not None and len(sedline) > 5:
                        tup7 = self.tuples_all_from_key(target_key)
                        if tup7 is not None and len(tup7) > 5:
                            linenum = tup7[6]
                            make_return = self.sedqueue(linenum,sedline)
                return make_return


            def toggle_safe(self,target_key,target_value=None,
                            safe=True,autoquote_level=0):
                """ If you have manually checked the existing setting then
                you might call this function with just a 'target_key' argument.
                For safer usage (better catching of logic / user errors) use
                two or more arguments.
                False will be returned if target_value is supplied and it is not
                in ['0','1','false','true','False','True','off','on','no','yes']
                False will be returned if we are safe=True and the toggle action
                would have no effect.

                  Example1 Loose: toggle_safe('X11Forwarding')
                  Example2 Safe: toggle_safe('PermitRootLogin','yes')
                  Example2 in words might be expressed as flick PermitRootLogin
                to 'yes' only if it is currently present and detected as 'no'
                Example3 Less Loose: toggle_safe('X11Forwarding','yes',False)
                In example 3 we are specifically telling the function that X11Forwarding
                should be set to 'yes' but if it is already 'yes' then function still
                returns True (do not treat 'leave value unchanged' as an error)

                Least error checking and most correcting way of invoking this function:
                  toggle_safe(READ_ENV,'yes',False,2) (says safe=False and autoquote_level=2)
                """
                toggle_return = True
                if target_key not in self:
                    """ Nothing we can toggle so do no more """
                    toggle_return = False
                elif target_value is not None:
                    """ Do some prechecking before possible calling sedline_toggle_wrapper """
                    stored_value = self[target_key]
                    tuple3before = self.false_true_index(stored_value)
                    tuple3after = self.false_true_index(target_value)
                    #print tuple3before,tuple3after
                    if tuple3after[0] is None:
                        toggle_return = False
                    else:
                        if tuple3after[0] == tuple3before[0]:
                            if safe:
                                """ When 'safe' treat 'leave value unchanged' as an error """
                                toggle_return = False
                            else:
                                toggle_return = True
                        else:
                            enclosed_by_before = self.enclosed_by_char(stored_value)
                            enclosed_by_after = self.enclosed_by_char(target_value)
                            if enclosed_by_after is None:
                                if enclosed_by_before is None:
                                    toggle_return = self.sedline_toggle_wrapper(target_key)
                                elif autoquote_level > 0:
                                    """ autoquote_level > 0 includes automatic quoting if
                                    quoting ommitted so proceed with toggle call """
                                    toggle_return = self.sedline_toggle_wrapper(target_key)
                                else:
                                    toggle_return = False
                            elif enclosed_by_after == enclosed_by_before:
                                toggle_return = self.sedline_toggle_wrapper(target_key)
                            elif autoquote_level > 1:
                                """ autoquote_level=2 includes automatic correction of
                                wrong quotes supplied by user so proceed with toggle call """
                                toggle_return = self.sedline_toggle_wrapper(target_key)
                            else:
                                toggle_return = False
                else:
                    """ No target_value supplied so just call sedline_toggle_wrapper() """
                    toggle_return = self.sedline_toggle_wrapper(target_key)
                return toggle_return


            def sed_checksum(self,sedlines,reverse=True,debug_filename=None,shell=None):
                """ Put the sedlines and original lines into a script and execute remotely
                to produce a readable result and checksum (md5)
                By default the original lines will be reversed so that the sed lines are
                simulating what happens if seds applied by working up from bottom of original
                file [lines]. Call function with reverse=False to not have reversed(),
                but this should not be the default.
                Returns a 3 tuple:
                (.) The lines of the script
                (.) The md5sum of the result
                (.) The lines of the result
                """
                scriptlines = []
                lines = []
                if sedlines is None or len(sedlines) < 1:
                    return (scriptlines,None,lines)
                tempfile_res = False

                scriptlines = ['#!/bin/dash\n']
                md5after_estimate = cmdbin.md5_from_iterable(self.lines_for_put())
                estimate_line = "# Forward estimate: {0} -> {1}\n".format(self.md5sum,md5after_estimate)
                scriptlines.append(estimate_line)
                if len(sedlines) > 1:
                    sed_cmdline = "{0} '{1}'".format(cmdbin.CMD_SEDE,' \\'.join(sedlines))
                else:
                    sed_cmdline = "{0} '{1}'".format(cmdbin.CMD_SEDE,sedlines[0])
                scriptlines.append(sed_cmdline)
                if reverse:
                    scriptlines.append(" << 'ORIGINALENDMARK' | {0}\n".format(cmdbin.CMD_TAC))
                    #scriptlines.append(" << 'ORIGINALENDMARK' | {0} 1>&2 \n")
                else:
                    scriptlines.append(" << 'ORIGINALENDMARK'\n")
                original_lines = self.lines_tuple
                if len(original_lines) < 1:
                    return (scriptlines,None,lines)
                if reverse:
                    scriptlines.extend(reversed(original_lines))
                else:
                    scriptlines.extend(original_lines)
                scriptlines.append('ORIGINALENDMARK')

                if debug_filename is not None:
                    if set(debug_filename).issubset(cmdbin.SET_PRINTABLE):
                        debug_filepath = "/tmp/{0}.dash".format(debug_filename)
                        with open(debug_filepath,'w') as f:
                            for line in scriptlines:
                                f.write(line)

                temp_res = False
                script_temp = tempfile.NamedTemporaryFile(delete=False)
                try:
                    for sline in scriptlines:
                        script_temp.write(sline)
                    script_temp.flush()
                    # script_temp.seek(0) # seek(0) so later read() from start
                    script_temp.close()
                except Exception as e:
                    raise
                finally:
                    temp_res = True
                if not temp_res:
                    return (scriptlines,None,lines)

                if shell is not None and shell.endswith('sh'):
                    sed_cmdline = "/bin/{0} {1}".format(shell,script_temp.name)
                else:
                    sed_cmdline = "/bin/dash {0}".format(script_temp.name)
                sed_rc,lines = cmdbin.local_popen_split_retaining(sed_cmdline)
                unlink(script_temp.name)
                #print sed_rc,len(lines)
                if sed_rc == 0:
                    pass
                else:
                    return (scriptlines,None,lines)
                mdf = cmdbin.md5_from_iterable(lines)
                if mdf is None:
                    return (scriptlines,None,lines)
                #print cmdbin.md5dict_from_iterable(lines)
                return (scriptlines,mdf,lines)


        list_of_tuples = []
        """ list_of_tuples will often be shorter in length than list_of_tuples_all
        as list_of_tuples_all will have an entry for EVERY LINE IN THE FILE
        Blank lines and other lines which are not able to be allocated a
        recognisable key will appear in list_of_tuples_all with first field of None
        """
        list_of_tuples_all = []
        linecount = -1
        lines = []
        """ When populating list_of_tuples_all we store the line unchanged
        and including any line separator. That way we have a high fidelity
        record of the original file content in a convenient iterable and
        this should allow full reconstruction of the file if desired.
        ( Why retain the line separator? We could attempt to strip based on
        os.linesep however this Python code is running locally but the file
        has been pulled from a remote system. )
        By contrast in list_of_tuples the second field will be stripped of spaces
        for convenient lookup of an individual value, but would not
        allow full reconstruction of the file.
        Use what is best depending on your purpose at the time.
        """

        keysep = None
        """ Most of the detailed parsing is handed off to 3 class methods which can
        be overridden in a subclass to suit your particular need. 3 class methods are:
        (.) process_comment_line
        (.) process_promising_line
        (.) process_line
        comment_line method returns a 2 tuple, other methods return a 4 tuple
        Third field of any 4 tuple is keysep
        Fourth field (keyd) allows for differences in stripping between
        the values that will be subsequently stored in list_of_tuples_all and list_of_tuples
        The fourth field will be the dictionary key used to populate the underlying OrderedDict
        """
        for line in self._given_lines:
            linecount += 1
            stripped = line.strip()
            stripped_len = len(stripped)
            if stripped_len < 1:
                try:
                    list_of_tuples_all.append((None,line,0,None,0,line[-1],linecount))
                except IndexError:
                    list_of_tuples_all.append((None,line,0,None,0,None,linecount))
            elif stripped_len == 1 and stripped == chr(35):
                """ Comment marker and nothing else """
                list_of_tuples_all.append((None,line,0,None,0,line[-1],linecount))
            elif stripped_len > self.MAX_LINELENGTH:
                """ MAX_LINELENGTH class variable accessed via self.MAX_LINELENGTH """
                raise Exception('MAX_LINELENGTH limit exceeded. Limited to prevent memory exhaustion')
            else:
                """ Examine the line further. Line will almost certainly make it into list_of_tuples """
                if stripped.split()[0] == chr(35):
                    """ Comment line """
                    key,value = (self.__class__).process_comment_line(stripped,keysep)
                    """ For comment lines no need for calling tuple7, just construct manually """
                    list_of_tuples_all.append((key,line,0,None,0,line[-1],linecount))
                    list_of_tuples = (self.__class__).list_of_tuples_appended(key,value,list_of_tuples)
                    #list_of_tuples.append((key,value))
                elif keysep is None:
                    """ Not a comment line. No keysep set yet """
                    if (self.__class__).KEYSEP_NAMED is None:
                        """ Usually the case unless we have subclassed """
                        keyfull,value,keysep,keyd = (self.__class__).process_promising_line(stripped)
                        """ We WILL OVERRIDE the existing keysep using third field returned """
                        list_of_tuples_all.append(self.tuple7(keyfull,line,linecount,keysep))
                        """ keyfull is used above, keyd is used below and is the key to go to OrderedDict """
                        list_of_tuples = (self.__class__).list_of_tuples_appended(keyd,value,list_of_tuples)
                        #list_of_tuples.append((keyd,value))
                    else:
                        """ KEYSEP_NAMED is set so accept the instruction and make keysep equal to
                        KEYSEP_NAMED. No need to call process_promising_line() for guesswork,
                        instead act appropriately now we know what keysep is to be. """
                        keysep = (self.__class__).KEYSEP_NAMED
                        if keysep in stripped:
                            keyfull,value,keysep2,keyd = (self.__class__).process_line(stripped,keysep)
                            list_of_tuples_all.append(self.tuple7(keyfull,line,linecount,keysep))
                            """ keyfull is used above, keyd is used below and is the key to go to OrderedDict """
                            list_of_tuples = (self.__class__).list_of_tuples_appended(keyd,value,
                                                                                      list_of_tuples)
                            #list_of_tuples.append((keyd,value))
                        else:
                            """ key=stripped and value=stripped for our purposes """
                            list_of_tuples_all.append((stripped,line,0,None,0,line[-1],linecount))
                            list_of_tuples = (self.__class__).list_of_tuples_appended(stripped,stripped,
                                                                                      list_of_tuples)
                            #list_of_tuples.append((key,value))
                elif (self.__class__).KEYSEP_NAMED is None or keysep in stripped:
                    """ keysep is not None. Likely a file with keys and vals separated by =/: or space """
                    keyfull,value,keysep2,keyd = (self.__class__).process_line(stripped,keysep)
                    """ We will not override the existing keysep so third field goes to keysep2
                    Fourth field above (keyd) is a mechanism for having a different (perhaps stripped?)
                    key added to list_of_tuples(). First field will be added to list_of_tuples_all either
                    directly or via tuple7 (see below)
                    """
                    list_of_tuples_all.append(self.tuple7(keyfull,line,linecount,keysep))
                    """ keyfull is used above, keyd is used below and is the key to go to OrderedDict """
                    list_of_tuples = (self.__class__).list_of_tuples_appended(keyd,value,
                                                                              list_of_tuples)
                    #list_of_tuples.append((keyd,value))
                else:
                    """ key=stripped and value=stripped for our purposes """
                    list_of_tuples_all.append((stripped,line,0,None,0,line[-1],linecount))
                    list_of_tuples = (self.__class__).list_of_tuples_appended(stripped,stripped,
                                                                              list_of_tuples)
                    #list_of_tuples.append((key,value))

            if linecount > self.MAX_LINES:
                """ MAX_LINES class variable accessed via self.MAX_LINES """
                raise Exception('MAX_LINES limit exceeded. Limited to prevent memory exhaustion')

        self._keyvalfile = Keyvalfile(list_of_tuples)
        """ for conf_k,conf_v in self._keyvalfile.items():
            if conf_k in ['SHELL','#SHELL']:
                print conf_v """
        self.keyvalfile.linecount_and_lines_set(self._given_linecount,self._given_lines,
                                                list_of_tuples_all,keysep)
        return self.keyvalfile


    def __exit__(self, extype, exvalue, extraceback):
        """ Int this context __exit__ function we use update_flag being False
        as a 'no further processing' marker.
        """
        exit_return = False
        """ Context Manager __exit__ returning False mean propagate the error """
        update_flag = True
        if self._bufio is not None:
            try:
                self._bufio.close()
            except Exception:
                """ If we fail here then assume any update attempt
                would be inadvisable so set update_flag False
                """
                update_flag = False

        if extype is None:
            if update_flag and self.keyvalfile.change_count > 0:
                """ When using 'put' update method we need to perform update here """
                pass
            elif self.keyvalfile.change_count > 0:
                warntail = "but no changes actioned!"
                warnline = "File {0} had {1} changes pending {2}".format(self.filepath,
                                                                         self.keyvalfile.change_count,
                                                                         warntail)
                print warnline
        else:
            from exceptions import AssertionError
            if extype == AssertionError:
                """ Test suite failures typically """
                exline = "Exception at context exit: {0}".format(exvalue)
                print exline
            else:
                exline = "File {0} encountered a {1} exception at context exit!".format(self.filepath,
                                                                                        extype)
                print exline
                print exvalue
                tbline = None
                try:
                    from traceback import format_exc
                    tbline = format_exc()
                except:
                    pass
                if tbline is not None:
                    exit_return = True
                    """ We have provided all possible information about the exception,
                    so instead of propagating it, we set above True """
                    print tbline
        return exit_return


        if update_flag and self.keyvalfile.change_count > 0:
            """ For both 'put' and 'sed' update methods we consult
            dict_of_changed to check length tallies with change_count.
            Regardless of update method, dict_of_changed and dict_of_seds
            are kept in step remember. """
            if self._change_count == len(self.dict_of_seds):
                update_flag = True
            else:
                update_flag = False
        else:
            update_flag = False

        if update_flag and self.change_method == 'put':
            """ Apply changes internaly and then use 'put' """
            print self.keyvalfile.lines()
        elif update_flag and self.change_method == 'sed':
            """ Sed method of updating using a /usr/bin/tac output of original file
            involves cmdbin.TAC -> cmdbin.SEDNF -> cmdbin.TAC """
            pass
        elif update_flag == False:
            exit_return = True
        return exit_return


class Keyedlarge(Keyedfile):
    """ Extension of parent Keyedfile for specialisation.
    If you have a configuration file with very long lines then the
    extra allowance in MAX_LINELENGTH might suit your needs.

    If you know the character that separates key from value
    for sure then you might instead use subclasses Keyedcolon
    and/or Keyedequals which in addition to working with
    given knowledge of what separates key from value
    also have MAX_LINES set 4000.

    Is a file having more than 4000 lines really a configuration file?
    """

    MAX_LINELENGTH = 1000
    MAX_LINES = 4000
    """ Above class variables are defined once for the class, but in practice
    accessed via self.MAX...
    The intention is to have a hard limit across the class, to help prevent
    accidental reading of things that might lead to memory exhaustion.
    But if you override or redefine things at runtime, then that is at your risk.
    """

    def __init__(self, *args, **kwargs):
        super(Keyedlarge, self).__init__(*args, **kwargs)


class Keyedcolon(Keyedfile):
    """ Extension of parent Keyedfile for specialisation.
    chr(58) is colon (:)            ; chr(61) is equals (=)
    Use this class to force key value separation by colon (:)
    """

    MAX_LINES = 4000
    KEYSEP_NAMED = chr(58)

    def __init__(self, *args, **kwargs):
        super(Keyedcolon, self).__init__(*args, **kwargs)


class Keyedequals(Keyedfile):
    """ Extension of parent Keyedfile for specialisation.
    chr(58) is colon (:)            ; chr(61) is equals (=)
    Use this class to force key value separation by colon (:)
    """

    MAX_LINES = 4000
    KEYSEP_NAMED = chr(61)

    def __init__(self, *args, **kwargs):
        super(Keyedequals, self).__init__(*args, **kwargs)


class Keyedequals2(Keyedfile):
    """ Extension of parent Keyedfile for specialisation.
    chr(58) is colon (:)            ; chr(61) is equals (=)
    Use this class to force key value separation by equals (=)
    Two techniques are available for forcing keysep to be
    assumed. The first is for a class to set KEYSEP_NAMED
    (recommended). The second (used here) is to override
    process_promising_line() which we do below.
    """

    def __init__(self, *args, **kwargs):
        super(Keyedequals2, self).__init__(*args, **kwargs)


    @classmethod
    def process_promising_line(cls,stripped):
        """ There is no keysep argument here as if keysep is known then you
        should probably instead be calling process_line().
        """
        # chr(58) is colon (:)            ; chr(61) is equals (=)           ; chr(32) is space
        keysep = chr(61)
        keyfull,value,keysep2,keyd = (self.__class__).process_line(stripped,keysep)
        return (keyfull,value,keysep,keyd)


class Keyedspace(Keyedfile):
    """ Extension of parent Keyedfile for specialisation.
    chr(58) is colon (:)            ; chr(61) is equals (=)           ; chr(32) is space
    Use this class to force key value separation by colon (:)
    """

    MAX_LINES = 4000
    KEYSEP_NAMED = chr(32)

    def __init__(self, *args, **kwargs):
        super(Keyedspace, self).__init__(*args, **kwargs)


def sed_remote_apply(md5current,filepath,sedlines=[],diffcheck='low',
                                 reverse=True,maxchanged=0.75):
    """ diffcheck low,medium,high is primarily about the use of diff tool
    but low,medium,high is used in other decisions so might be thought of
    as an operating mode also.

    'low' Simple approach to counting difference where diffutils is available.
    Our maxchanged implementation relies on presence of diffutils so when diffcheck='low' it may not
    be always respected. There is an md5sum check just before the final move.

    'medium' requires diffutils exist on the remote system. Medium approach to counting difference
    Will use maxchanged parameter and prevent replacement of file on remote system if the
    changes exceed maxchanged. As with 'low' there is an md5sum check.

    'high' High approach to counting difference.
    Will use maxchanged parameter and prevent replacement of file on remote system if the
    changes exceed maxchanged. As with 'low' and 'medium' there is an md5sum check.
    """
    update_return = False
    if diffcheck not in ['low','medium','high']:
        """ Treat bad value given for diffcheck as 'medium' diffcheck level """
        diffcheck = 'medium'
    if sedlines is None or len(sedlines) < 1:
        return False
    res_lines = []
    with settings(warn_only=True):
        version_res = cmdbin.run_or_sudo(cmdbin.CMD_DIFF_VERSION,True,True)
        res_lines = cmdbin.run_res_lines_stripped(version_res)
    diffutils_flag = False
    for line in res_lines:
        if 'diffutils' in line:
            diffutils_flag = True
    if diffutils_flag is False and diffcheck in ['medium','high']:
        return False
    md5sum = cmdbin.md5text_or_sudo(filepath)
    if md5sum is None:
        return False

    filepath_dir = path.dirname(filepath)
    filepath_base = path.basename(filepath)
    sed_diff_lines = []
    with cd(filepath_dir):
        sedfile = "{0}.sed".format(filepath_base)
        put_res = cmdbin.put_using_named_temp_sudo(sedfile,sedlines)
        if put_res:
            print "put of .sed was successful"
        else:
            return False
        sed_diff_line = "{0} {1} < {2} | {3} {4} - ;".format(cmdbin.CMD_SEDF,sedfile,filepath_base,
                                                      cmdbin.CMD_DIFF_UNIFIED_ZERO,filepath)
        """ Planned: cmdbin.CMD_TAC will be used when forming sed_diff_line for advanced use. """
        with settings(warn_only=True):
            sed_diff_res = cmdbin.run_or_sudo(sed_diff_line)
            if sed_diff_res.return_code == 1:
                from os import linesep
                sed_diff_lines = sed_diff_res.split(linesep)

    single_plus_lines = []
    for line in sed_diff_lines:
        if line.startswith(chr(43)) and not line.startswith('+++'):
            single_plus_lines.append(line)
    unified_count = len(single_plus_lines)
    if unified_count == 0:
        """ Diff could not detect any changes made by the sed """
        print 'Diff could not detect any changes made by the sed!'
        return False
    for line in single_plus_lines:
        print line
    # For medium we might do some extra checking using cmdbin.CMD_DIFF_SIDEBYSIDE_NOCOMMON
    # sed_remote_apply() is a prototype and more work is needed especially for 'medium' and 'high'
    update_return = True
    return update_return


def useradd_shell(preferred='/bin/sh'):
    """ Returns a three tuple of 'now',changed,'before'
    Example Return ('/bin/bash',True,'/bin/zsh')
    Above indicates that '/bin/bash' is the current value,
    it was previously '/bin/zsh' and the True indicates that we changed.
    /bin/sh is often the default on Debian,
    whereas on CentOS specifically set as /bin/bash
    """
    """ This is a dummy implementation as augeas relies on predefined templates
    so probably overkill for general (non-specific) key value style files """
    shell_now = preferred
    shell_original = preferred
    """
    from augeas import augeas
    aug = augeas()
    shell_original = aug.get('/etc/default/useradd/SHELL')
    shell_now = shell_original
    """
    return (shell_now,False,shell_original)

