#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2013 Gary Wright wrightsolutions.co.uk/contact
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# Tested against fabric 1.4.3
# http://ftp.uk.debian.org/debian/pool/main/f/fabric/fabric_1.4.3-1.debian.tar.gz

""" DNS specifically and anything related to lookups / resolutions.

Note: Many of the configuration files related to dns are very specific
formats here are two things which mark them out as variants or special:
(.) options allows multiple values whereas generic Keyedfile is really one key -> one value
(.) options allows values that are themselves delimited Example: options timeout:2
See augeas for a pre templated solution to parsing non ini configuration files
See Keyedcolon() subclass of conf_get_set for a key value parser that forces colon (:)
as separator.

For the nameserver functions we use keylines_typed() rather than
straight dictionary lookups because the underlying OrderedDict is
designed to keep non-duplicates (in the dict) and we know
and expect [enabled] duplicates in our resolv.conf (Example: two nameserver entries)

To access two [enabled] lines shown below via a "duplicates replace" key value api
we use query the underlying data using keylines_typed(1,...) instead of straight 'in':
domain cable.buried.net
search cable.buried.net buried.net
nameserver 8.8.8.8
nameserver 8.8.4.4
options attempts:3
options timeout:2
options rotate
options inet6
#options single-request
options single-request-reopen
# From manpage of resolv.conf:
# The keyword and value must appear on a single line,
# and the keyword (e.g., nameserver) must start the line.
# The value follows the keyword, separated by white space.
# Any problems with 'options rotate' see following test
# python -c 'from socket import gethostbyname;print [ "Y" if gethostbyname("debian.org") else "." for x in range(10) ]'
"""



from __future__ import with_statement
from fabric.api import abort, cd, env, get, put, run, settings, sudo
#from fabric.contrib.console import confirm
from fabric.contrib.files import exists,append
from os import path
import re
from . import cmdbin
from . import conf_get_set as getset

CMD_RESCONF="/sbin/resolvconf "
CMD_RESCONF_UPDATE="/sbin/resolvconf -u "

RESCONF='/etc/resolv.conf'
RESCONF_DIR='/etc/resolvconf'

GOOGLE4LIST=['8.8.8.8','8.8.4.4']
GOOGLE6LIST=['2001:4860:4860::8888','2001:4860:4860::8844']
GOOGLELIST=GOOGLE4LIST[:]
GOOGLELIST.extend(GOOGLE6LIST)


def nameserver_count():
    count = 0
    with getset.Keyedfile(RESCONF) as conf:
        for line in conf.keylines_typed(1,['nameserver']):
            count += 1
    return count


def nameserver_count_excluding(exclude_list,print_flag=False):
    count = 0
    COUNT_PREFIX = 'Nameserver Count (excluding)'
    if hasattr(exclude_list, "__iter__"):
        pass
    elif exclude_list == 'google':
        exclude_list = GOOGLELIST[:]
    elif print_flag == True:
        print "{0} is {1} (counting terminated early)".format(COUNT_PREFIX,count)
        return count
    else:
        return count
    with getset.Keyedfile(RESCONF) as conf:
        for line in conf.keylines_typed(1,['nameserver']):
            line_array = line.strip().split()
            if line_array[1] in exclude_list:
                pass
            else:
                count += 1
    if print_flag == True:
        print "{0} is {1}".format(COUNT_PREFIX,count)
    return count


def nameserver_count_startswith(starts,print_flag=False,verbose=False):
    count = 0
    COUNT_PREFIX = 'Nameserver Count (startswith)'
    with getset.Keyedfile(RESCONF) as conf:
        for line in conf.keylines_typed(1,['nameserver']):
            line_array = line.strip().split()
            if line_array[1].startswith(starts):
                count += 1
            else:
                pass
    if print_flag == True:
        if count > 0 or verbose == True:
            print "{0} is {1}".format(COUNT_PREFIX,count)
    return count
    

def resolvconf_dirs_more(main_itself=False):
    more_res = False
    #which_res = run("{0} {1}".format(cmdbin.CMD_WHICH,'resolvconf'))
    exists_resconf = exists(CMD_RESCONF,verbose=True)
    if exists_resconf:
        ros_line = "{0} {1}/{2}/* {1}/{3}/*".format(cmdbin.CMD_FGREP_H_V_Z,
                                                    RESCONF_DIR,
                                                    'resolv.conf.d',
                                                    'update.d')
        ros_res = cmdbin.run_or_sudo(ros_line)
        if ros_res.succeeded:
            more_res = True
    return more_res

