#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#   Copyright 2013 Gary Wright wrightsolutions.co.uk/contact
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

# Tested against fabric 1.4.3
# http://ftp.uk.debian.org/debian/pool/main/f/fabric/fabric_1.4.3-1.debian.tar.gz

from __future__ import with_statement
from fabric.api import abort, cd, env, get, hide, put, run, settings, sudo
from fabric.contrib.console import confirm
from fabric.contrib.files import exists,append
from os import path
import re
from . import cmdbin

""" res.succeeded is preferred for newer fabric, but not res.failed
is used where backward compatibility is preferable. """

CMD_APT = "/usr/bin/apt-get "
CMD_APT_NORECOMMENDS = "/usr/bin/apt-get --no-install-recommends "
CMD_APT_ASSUMEYES = "/usr/bin/apt-get --assume-yes "
CMD_APT_ASSUMEYES_NORECOMMENDS = "/usr/bin/apt-get --no-install-recommends -y "
CMD_APT_ASSUMEYES_NORECOMMENDS_CONFOLD = "/usr/bin/apt-get \
 -o DPkg::Options::=--force-confold --no-install-recommends -y "
CMD_APT_UPDATE = "/usr/bin/apt-get update "
CMD_DPKG = "/usr/bin/dpkg "
CMD_DPKG_STATUS = "/usr/bin/dpkg -s "
CMD_DPKG_QUERY_STATUS = "/usr/bin/dpkg-query -W --showformat='${Status;-30}'"
CMD_DPKG_QUERY_STATUS4 = "dpkg-query -W \
--showformat='${Installed-Size;-10}\t${Status;-30}\t${Section;-15}\t${Package;-30}\n' "
CMD_NETSELECT_APT = "/usr/bin/netselect-apt"

RE_STATUS_INSTALLED = re.compile('Status:.*installed')
RE_NO_PACKAGES_MATCHING = re.compile(r'no\spackages\sfound\smatching\b')
RE_OK_INSTALLED = re.compile(r'\bok\sinstalled\b')

_list_dot_d_array = []
_sources_list_object = None


def list_dot_d_array_extend(extend_with):
    global _list_dot_d_array
    extended = True
    try:
        _list_dot_d_array.extend(extend_with)
    except Exception,e:
        extended = False
    return extended


def get_list_dot_d_array():
    return _list_dot_d_array


def has_backports():
    has_flag = False
    with settings(warn_only=True):
        REGEX = '.*(/sources.list|/sources.list.d/.*\.list)'
        find_line = "{0} /etc/apt/ {1} {2}{3}{2} -exec cat {4}".format(cmdbin.CMD_FIND,
                                                                       cmdbin.REGEX_OPTS_EGREP,
                                                                       chr(39),REGEX,cmdbin.EXEC_TAIL)
        ros_res = cmdbin.run_or_sudo(find_line)
        res_lines = cmdbin.run_res_lines_stripped(ros_res)
        for line in res_lines:
            if line.startswith('deb '):
                if '-backports ' in line:
                    print "MATCH: {0}".format(line)
                    has_flag = True
        """ find /etc/apt/ -regextype posix-egrep -regex \
        '.*(/sources.list|/sources.list.d/.*\.list)' -exec cat {} \;
        Above is a longhand representation of our find_line construction.
        What follows is okay but will bork when nothing in sources.list.d
        grep_line = "{0} '^deb ' {1}/sources.list {1}/sources.list.d/*.list".format(
            cmdbin.CMD_EGREP,'/etc/apt') """
    return has_flag


def dpkg_status_as_list(packagename=None):
    if not packagename:
        return []

    with settings(warn_only=True):
        """ cmdbin.run_or_sudo("{0} {1}".format(CMD_DPKG_STATUS,packagename)) """
        status_res = run("{0} {1}".format(CMD_DPKG_STATUS,packagename))
        cmdbin.print_success_and_failed(status_res)

    status_list = []
    if status_res.succeeded:
        status_list = [ ln for ln in status_res.split('\n') \
                            if RE_STATUS_INSTALLED.match(ln) ]

    return status_list


def dpkg_status_installed(packages_list=[],return_type='flag'):
    """ By default this function returns True or False,
    but giving second argument of 'outstanding' you would
    instead have a list returned.
    """
    packages_list_count = len(packages_list)
    if packages_list_count < 1:
        return False
    installed_list_count = 0
    installed_res = True
    outstanding_list = []
    for package in packages_list:
        if set(package).issubset(cmdbin.SET_PRINTABLE):
            res_lines = []
            with settings(warn_only=True):
            #with hide('output'):
                status_res = run("{0} {1}".format(CMD_DPKG_QUERY_STATUS4,package))
                res_lines = cmdbin.run_res_lines_stripped(status_res)
            if status_res.succeeded:
                for line in res_lines:
                    line_search = line[:42]
                    if RE_NO_PACKAGES_MATCHING.search(line_search):
                        outstanding_list.append(package)
                    elif RE_OK_INSTALLED.search(line_search):
                        installed_list_count += 1
                    else:
                        outstanding_list.append(package)
            else:
                outstanding_list.append(package)
        else:
            installed_res = False
            break
    print len(outstanding_list)
    if installed_res and installed_list_count == packages_list_count:
        installed_res = True
        outstanding_list = []
    if return_type == 'outstanding':
        return outstanding_list
    return installed_res


def source_list_lines_debian(pretty_dict,repolabel,backports_main_only=True):
    source_list_lines = None
    if repolabel == 'backports' and backports_main_only:
        source_list_lines = """
# added by fabutil
deb http://ftp.debian.org/{id} {release}-backports main
deb-src http://ftp.debian.org/{id} {release}-backports main
""".format(**pretty_dict)
    elif repolabel == 'backports':
        source_list_lines = """
# added by fabutil
deb http://ftp.debian.org/{id} {release}-backports main contrib non-free
deb-src http://ftp.debian.org/{id} {release}-backports main contrib non-free
""".format(**pretty_dict)
    elif repolabel in ['contrib','non-free']:
        source_list_lines = """
# added by fabutil
deb http://http.debian.net/{id}/ {release} {repolabel}
deb-src http://http.debian.net/{id}/ {release} {repolabel}
""".format(**pretty_dict)
    else:
        pass
    return source_list_lines


def sources_list_lines(list_dot_d):
    sl_lines = None
    (pname, pretty_name_id, pretty_name_word) = cmdbin.pretty_name_tuple()
    if pretty_name_id == 'debian':
        """ See www.debian.org/doc/debian-policy/ch-archive.html for more about
        distinction between main and other repositories
        """
        if list_dot_d in ['backports','contrib','non-free']:
            pretty_dict = {'id': pretty_name_id,
                           'release': pretty_name_word,
                           'repolabel': list_dot_d}
            sl_lines = source_list_lines_debian(pretty_dict,list_dot_d)
        else:
            pass
    elif pretty_name_id == 'ubuntu':
        """ Unlike Debian which distinguishes repos by 'freedom to modify/distribute'
        Ubuntu distinguishes repos by level of support by Canonical company.
        A firm grasp on this distinction will help you avoid simplistic maps, but next
        is one to get you started: (contrib+non-free)->company filter->(multiverse+restricted)
        """
        if list_dot_d in ['universe','multiverse']:
            pretty_dict = {'id': pretty_name_id,
                           'release': pretty_name_word,
                           'verse': list_dot_d}
            sl_lines = """
# added by fabutil
deb http://archive.ubuntu.com/{id}/ {release} {verse}
deb-src http://archive.ubuntu.com/{id}/ {release} {verse}
""".format(**pretty_dict)
    else:
        pass
    return sl_lines


def aptget_line(packages_list=[],assumeyes=False,
                recommends=True,confold=False,precheck=False):

    """ Returning an empty string indicates nothing to do (no packages need installing)
    For exceptions/unexpected situations then raise Exceptions
    """

    packages_list_count = len(packages_list)
    if packages_list_count < 1:
        raise Exception('Unexpected value as first argument')

    if precheck:
        packages_list = dpkg_status_installed(packages_list,'outstanding')
        if len(packages_list) < 1:
            return ''

    packages = ' '.join(packages_list)
    apt_line = "{0} install {1}".format(CMD_APT,packages)
    if assumeyes:
        if recommends is False:
            if confold:
                apt_line = "{0} install {1}".format(CMD_APT_ASSUMEYES_NORECOMMENDS_CONFOLD,
                                                    packages)
            else:
                apt_line = "{0} install {1}".format(CMD_APT_ASSUMEYES_NORECOMMENDS,
                                                    packages)
        else:
            apt_line = "{0} install {1}".format(CMD_APT_ASSUMEYES,packages)
    else:
        if recommends is False:
            apt_line = "{0} install {1}".format(CMD_APT_NORECOMMENDS,packages)
    return apt_line


def requires_packages(packages_list=[],assumeyes=False,
                      recommends=True,confold=False):
    """ Installs required packages.
    Unlike the 'one by one' variant requires_dpkg_package() this function
    has no list_dot_d argument. So if you did need to install from
    'contrib', 'non-free', 'universe', 'multiverse' and you want automatic
    repository adding, then instead use requires_dpkg_package()
    """
    requires_res = False
    ros_line = aptget_line(packages_list,assumeyes,recommends,confold,True)
    ros_res = cmdbin.run_or_sudo(ros_line)
    if ros_res.succeeded:
        requires_res = True  
    return requires_res


def requires_dpkg_package(packagename=None,assumeyes=False,
                          recommends=True,confold=False,list_dot_d=None):
    if packagename:
        requires_res = True
    else:
        return False

    if len(dpkg_status_as_list(packagename)) > 0:
        return True

    requires_res = False

    with settings(warn_only=True):
        apt_line = "{0} install {1}".format(CMD_APT,packagename)
        if assumeyes:
            if recommends is False:
                if confold:
                    apt_line = "{0} install {1}".format(CMD_APT_ASSUMEYES_NORECOMMENDS_CONFOLD,
                                                        packagename)
                else:
                    apt_line = "{0} install {1}".format(CMD_APT_ASSUMEYES_NORECOMMENDS,
                                                        packagename)
            else:
                apt_line = "{0} install {1}".format(CMD_APT_ASSUMEYES,packagename)
        else:
            if recommends is False:
                apt_line = "{0} install {1}".format(CMD_APT_NORECOMMENDS,packagename)
        ros_res = cmdbin.run_or_sudo(apt_line)
        if ros_res.succeeded:
            requires_res = True  
        elif list_dot_d:
            sl_name = "{0}.list".format(list_dot_d)
            sl_lines = None
            (pname, pretty_name_id, pretty_name_word) = cmdbin.pretty_name_tuple()
            if pretty_name_id == 'debian':
                """ See www.debian.org/doc/debian-policy/ch-archive.html for more about
                distinction between main and other repositories
                """
                if list_dot_d in ['contrib','non-free']:
                    pretty_dict = {'id': pretty_name_id,
                                   'release': pretty_name_word,
                                   'repolabel': list_dot_d}
                    sl_lines = """
# added by fabutil
deb http://http.debian.net/{id}/ {release} {repolabel}
deb-src http://http.debian.net/{id}/ {release} {repolabel}
""".format(**pretty_dict)
            elif pretty_name_id == 'ubuntu':
                """ Unlike Debian which distinguishes repos by 'freedom to modify/distribute'
                Ubuntu distinguishes repos by level of support by Canonical company.
                A firm grasp on this distinction will help you avoid simplistic maps, but next
                is one to get you started: (contrib+non-free)->company filter->(multiverse+restricted)
                """
                if list_dot_d in ['universe','multiverse']:
                    pretty_dict = {'id': pretty_name_id,
                                   'release': pretty_name_word,
                                   'verse': list_dot_d}
                    sl_lines = """
# added by fabutil
deb http://archive.ubuntu.com/{id}/ {release} {verse}
deb-src http://archive.ubuntu.com/{id}/ {release} {verse}
""".format(**pretty_dict)
            else:
                pass
            if sl_lines:
                put_flag = cmdbin.put_using_named_temp("/etc/apt/sources.list.d/{0}".format(sl_name),
                                                       sl_lines)
                if put_flag:
                    list_dot_d_array_extend(sl_name)

                    cmdbin.run_or_sudo(CMD_APT_UPDATE)

                    apt_line = "{0} install {1}".format(CMD_APT,packagename)
                    if assumeyes:
                        if recommends is False:
                            if confold:
                                apt_line = "{0} install {1}".format(CMD_APT_ASSUMEYES_NORECOMMENDS_CONFOLD,
                                                                    packagename)
                            else:
                                apt_line = "{0} install {1}".format(CMD_APT_ASSUMEYES_NORECOMMENDS,
                                                                    packagename)
                        else:
                            apt_line = "{0} install {1}".format(CMD_APT_ASSUMEYES,packagename)
                    else:
                        if recommends is False:
                            apt_line = "{0} install {1}".format(CMD_APT_NORECOMMENDS,packagename)
                    ros_res = cmdbin.run_or_sudo(apt_line)
                    if ros_res.succeeded:
                        requires_res = True
                else:
                    requires_res = False
                    """ Above else: for clarity rather than necessity """

    status_list = dpkg_status_as_list(packagename)
    print "{0} - {1}".format(packagename, status_list)
    # if len(dpkg_status_as_list(packagename)) > 0:
    if len(status_list) > 0:
        return True

    return requires_res


def requires_binary_package(binary_unpathed=None,packagename_suggestion=None,
                            assumeyes=False):
    """ The word 'binary' here refers to the fact that you supply a binary
    name as the first argument. Two examples follow:
    requires_res = requires_binary_package('ab','apache2-utils',assumeyes=True)
    requires_binary_package('django-admin','python-django',assumeyes=True)
    """
    which_res = None
    requires_which = cmdbin.CMD_KERN_RELEASE
    with settings(warn_only=True):
        which_res = run("{0} {1}".format(cmdbin.CMD_WHICH,binary_unpathed))

    print "{0} {1}".format(which_res.succeeded,which_res.failed)
    if which_res and which_res.succeeded:
        """ .succeeded is supported in Fabric 1.0 and newer """
        requires_which = which_res
        print which_res
        return requires_which
    elif which_res is not None and which_res.failed:
        if assumeyes:
            cmdbin.run_or_sudo("{0} install {1}".format(
                    CMD_APT_ASSUMEYES,packagename_suggestion))
        else:
            cmdbin.run_or_sudo("{0} install {1}".format(
                    CMD_APT,packagename_suggestion))

        with settings(warn_only=True):
            which_res = run("{0} {1}".format(cmdbin.CMD_WHICH,binary_unpathed))

        if which_res and not which_res.failed:
            requires_which = which_res
            print which_res
            return requires_which
    else:
        return which_res
    return False


def netselect_apt(verbose=False):
    netselect_return = False
    requires_res = requires_packages(['netselect','netselect-apt'],True,False,False)
    if verbose:
        ros_res = cmdbin.run_or_sudo(CMD_NETSELECT_APT)
    else:
        with hide('output'):
            ros_res = cmdbin.run_or_sudo(CMD_NETSELECT_APT)
    if ros_res.succeeded and ros_res.find('Writing sources.list'):
        netselect_return = True
        print 'A new sample sources.list has been generated (/root/sources.list?)'
        print 'Inspect and if necessary copy into /etc/apt/ the results you are happy with'
    return netselect_return



