#!/usr/bin/env python
# -*- coding: utf-8 -*-
from unittest import TestCase
from os import linesep
from string import ascii_lowercase,digits,printable,punctuation
from unicodedata import name as unicode_name

import cmdbin
import conf_get_set as getset

class KeyedLocal(TestCase):
    """ Key value files are not in themselves hard, but the problem of consistency
    and small differences in spacing / formatting or in the key separator used,
    make writing generic parsers a task involving consideration of many cases.

    If you want to understand why YAML was created then reading these tests and
    adding a handful of your own, should help understand the motivation for
    creating YAML and JSON and other conformant formats. YAML balances a
    lightweight parser with a few recommended guidelines on spacing / indentation
    """

    def setUp(self):
        """ Even if the application under testing does inline replaces,
        that will not affect some of these constants as they employ:
        (.) reversed() in definition
        (.) shallow copy and reverse just prior to test use. See [::-1]
        Some constants, particularly the multiline definitions do not
        employ this technique.
        """

        # chr(33) is exclamation mark (!)      ; chr(34) is double quote (")     ; chr(39) is single quote (')
        # chr(35) is hash / octothorpe             ; chr(58) is colon (:)            ; chr(61) is equals (=)

        self.SET_COLON_SPACE_EQUALS = set([chr(58),chr(32),chr(61)])
        self.SET_PRINTABLE = set(printable)
        self.SET_PUNCTUATION = set(punctuation)
        self.SET_LOWER_PLUS_DIGITS = set(ascii_lowercase).union(digits)

        self.APACHE2NAMEVIRTUALHOST = "".join(reversed('NameVirtualHost *:80{0}Listen 80{0}'.format(linesep)))
        """ Next is also lifts selected items/ideas from an Apache file
        and is based on a 2 section selective extract from a typical ports.conf """
        self.PORTS2SECTION = """
<IfModule mod_ssl.c>
    # If you add NameVirtualHost *:443 here, you will also have to change
    # the VirtualHost statement in /etc/apache2/sites-available/default-ssl
    # to <VirtualHost *:443>
    Listen 443
</IfModule>

<IfModule mod_gnutls.c>
    #Listen 443
    Listen 44443
</IfModule>
"""
        self.PORTS2SECTION_SPLIT = self.PORTS2SECTION.split(linesep)
        self.PORTS1LISTEN = "".join(reversed('Listen 80{0}'.format(linesep)))
        self.PORTS1LISTENAST = "".join(reversed('Listen *:80{0}'.format(linesep)))
        self.PORTS1LISTENLOCAL = "".join(reversed('Listen 127.0.0.1:80{0}'.format(linesep)))
        self.CRON_READ_ENV = "".join(reversed('#{0}READ_ENV="yes"{0}#{0}'.format(linesep)))
        self.TCPKEEPALIVEPLUSPLUS = r"""TCPKeepAlive yes
AcceptEnv LANG LC_*
Subsystem sftp /usr/lib/openssh/sftp-server
"""
        self.TCPKEEPALIVEPLUSPLUSSPLIT = self.TCPKEEPALIVEPLUSPLUS.split(linesep)
        """ Next we simulate some lines from /etc/default/useradd with a final hacky example
        - perhaps it might have been manually edited in a hurry
        """
        self.USERADD_RPMBASED5 = "".join(reversed("GROUP=100{0}HOME=/home{0}INACTIVE=-1{0}EXPIRE={0}SHELL=/bin/bash".format(linesep)))
        self.USERADD_RPMBASED5HACKY = "".join(reversed("GROUP=100{0}HOME=/home{0}INACTIVE=-1{0}EXPIRE={0}SHELL = /bin/sh{0}#SHELL=/bin/tcsh{0}# SHELL=/bin/zsh4{0}SHELL  =/bin/bash".format(linesep)))
        self.TRAFFIC3OFF = """
red:off
orange:off
green:off
"""
        self.TRAFFIC3OFF_SPLIT = self.TRAFFIC3OFF.split(linesep)
        self.TRAFFIC3GREEN = """
green:off
green:off
green:on
"""
        self.TRAFFIC3GREEN_SPLIT = self.TRAFFIC3GREEN.split(linesep)
        """ Simulating real life would see a requirement for one colour to be 'on' and the
        other two to be 'off' however we are simply treating each as an individual key value
        and do not implement any interdependency - traffic lights colours just provide
        a convenient example of a file that has 3 key values separated by (in this case) colon (:)
        """
        self.TRAFFIC3HACKY = """
#red:on
red:on
red:off
#orange:on
#orange : on
# orange:on
#orange : on
#orange   :   on
orange:on
orange :on
orange :  off
orange:on
orange :on
orange :  off
green:off
"""
        self.TRAFFIC3HACKY_SPLIT = self.TRAFFIC3HACKY.split(linesep)

        """ Next we mix in a directive style set of indents as if traffic lights
        configuration suddenly inherited the IfModule features of Apache.
        We do this to document current behaviour and to illustrate how the default
        Keyedfile class would treat such a chimera.
        """
        self.TRAFFIC3MODULES = """
red:off
orange:off
<IfModule traffic.c>
    orange:on
</IfModule>
green:on
    green:off
"""
        self.TRAFFIC3MODULES_SPLIT = self.TRAFFIC3MODULES.split(linesep)

        """ Next we try and deliberately confuse any parser guesswork by inserting
        a bad directive at the start (borrowed from smb). Early lines are important
        to any guesswork mechanism so inconsistent use of key value separators
        is discouraged. Here we are testing so create some artificial inconsistency.
        """
        self.TRAFFIC3OFF_BAD = """
cups server = mycupsserver:1631
red:off
orange:off
green:off
"""
        self.TRAFFIC3OFF_BAD_SPLIT = self.TRAFFIC3OFF_BAD.split(linesep)
        """ Next we alter the above example to give first line that
        might be marked 'space separated' by guesswork
        """
        self.TRAFFIC3OFF_BAD2 = """
mycupsserver 1631
red:off
orange:off
green:off
"""
        self.TRAFFIC3OFF_BAD2_SPLIT = self.TRAFFIC3OFF_BAD2.split(linesep)
        """ Next we alter the above example to give first line that
        might be marked 'equals separated' by guesswork
        """
        self.TRAFFIC3OFF_BAD3 = """
    port        = 3306
red:off
orange:off
green:off
"""
        self.TRAFFIC3OFF_BAD3_SPLIT = self.TRAFFIC3OFF_BAD3.split(linesep)

        self.PHP5INI_SIMILAR = """[PHP]
;;;;;;;;;;;;;;;;;;;;;;;
; Not a real php5.ini ;
;;;;;;;;;;;;;;;;;;;;;;;
; conf_get_set.py does not aim to support this sort of file
; as there are already good solutions (configparser) for working with ini files
; But we include it in our tests as it provides good edge cases / counter examples

{0}PostgresSQL{1}
; Allow or prevent persistent links.
pgsql.allow_persistent = On

{2} Detect broken persistent links always with pg_pconnect().
{2} Auto reset feature requires a little overheads.
pgsql.auto_reset_persistent = Off

; Maximum number of persistent links.  -1 means no limit.
pgsql.max_persistent = {3}1

; Comment line containing an = (equals symbol)
""".format(chr(91),chr(93),chr(59),chr(45))

        """ Above mixes in chr() definitions so that any encoding issues will
        be highlighted in any manual inspection of the output
        """

        # chr(91) is open bracket ([)      ; chr(93) is close bracket (])
        # chr(59) is semicolon (;)             ; chr(45) is minus sign  (-)

        self.PHP5INI_SIMILAR_SPLIT = self.PHP5INI_SIMILAR.split(linesep)
        self.PHP5INI_FILLER = '; filler'
        self.PHP5INI_FILLER_NEWLINED = "; filler{0}".format(linesep)
        self.PHP5INI_SIMILAR_LARGE = self.PHP5INI_SIMILAR + self.PHP5INI_FILLER_NEWLINED*2000
        self.PHP5INI_SIMILAR_LARGE_SPLIT = self.PHP5INI_SIMILAR_LARGE.split(linesep)

        """ Next example borrow some aspects of the look and feel of a my.cnf MySQL configuration
        We are particularly interested in repetition so that we can test toggling and that the
        correct target lines are changed when we update using 'put' method or 'sed' method
        Note: We have ressurected an old mysql directive skip-networking and given it a
        toggle friendly value. The representation below borrows some aspects of a mysql cnf
        file but should not be considered a model / true representation.
        """
        self.MYSQL5CNF_SIMILAR = """
# The MySQL database server configuration file.
 
[client]
#    port        = 3
    #port        = 6
    port        = 3306
   port         = 336
     port       = 33366
#socket          = /var/run/mysqld/mysqld.sock
socket          = /var/run/mysqld/mysqld2.sock
#socket          = /var/run/mysqld/mysqld.sock

[mysqld_safe]
  socket          = /var/run/mysqld/mysqld2.sock
  skip-networking   = on
   skip-networking   = off

[mysqld]
socket          = /var/run/mysqld/mysqld.sock
   port         = 336
  skip-networking   = on
"""
        self.MYSQL5CNF_SIMILAR_SPLIT = self.MYSQL5CNF_SIMILAR.split(linesep)

        self.MERGETOOLSRC = """
meld.gui=True
meld.args=--label='local' $local --label='base' $base --label='other' $other
meld.diffargs=-a --label='$plabel1' $parent --label='$clabel' $child
tkdiff.args=$local $other -a $base -o $output
tkdiff.gui=True
tkdiff.priority=-8
tkdiff.diffargs=-L '$plabel1' $parent -L '$clabel' $child
xxdiff.args=--show-merged-pane --exit-with-merge-status --title1 local --title2 base --title3 other --merged-filename $output --merge $local $base $other
xxdiff.gui=True
xxdiff.priority=-8
xxdiff.diffargs=--title1 '$plabel1' $parent --title2 '$clabel' $child
"""
        self.MERGETOOLSRC_SPLIT = self.MERGETOOLSRC.split(linesep)

        self.MERGETOOLSDIACRITIC = """
# Code reference at wikipedia.org/wiki/Acute_accent has lowercase e acute tabulated
# unicode: U+00E9
# html: &#233;
# \xc3\xa9 LATIN SMALL LETTER E WITH ACUTE
# Verify by websearching for 'diacritic' and looking up in a result table
ecmerge.args=$base $local $other --mode=merge3 --title0=base --title1=local --title2=other --to=$output
ecmerge.regkey=Software\Elli\xc3\xa9 Computing\Merge
ecmerge.regkeyalt=Software\Wow6432Node\Elli\xc3\xa9 Computing\Merge
ecmerge.gui=True
ecmerge.diffargs=$parent $child --mode=diff2 --title1='$plabel1' --title2='$clabel'
"""
        self.MERGETOOLSDIACRITIC_SPLIT = self.MERGETOOLSDIACRITIC.split(linesep)
        return


    def test_keysep(self):
        lines = self.APACHE2NAMEVIRTUALHOST[::-1].split(linesep)
        """ This is a tricky example if it were to appear as the first line of a file
        as we are simulating here. Should the program identify space as a separator of
        key value or should it identify colon (:) as separator of key value? 
        """
        with getset.Keyedfile(lines) as conf:
            self.assertEqual(conf.keysep,' ')
            tup7 = conf.tuples_all_from_key('NameVirtualHost')
            self.assertEqual('*:80',conf['NameVirtualHost'],"Not equal '*:80'. Tuple:{0}".format(tup7))
        return


    def test_word_in_value(self):
        with getset.Keyedfile(self.TCPKEEPALIVEPLUSPLUSSPLIT) as conf:
            self.assertEqual(conf.linecount,4)
            #print conf.tuples_all
            #for conf_k,conf_v in conf.items():
            #    print "{0}...{1}".format(conf_k,conf_v)
            self.assertEqual(len(conf),3)
            """ An explanation of the previous two assertEqual tests follows:
            len(self.TCPKEEPALIVEPLUSPLUSSPLIT) is 4 so conf.linecount should be 4 also
            There are in fact only 3 key value pairs and as conf is an underlying OrderedDict
            we can simply len(conf) and should expect the answer 3
            """
            self.assertTrue('AcceptEnv' in conf)
            """ Next we expect the character count for LANGLC_* (no spacing) to be 8 """
            self.assertEqual(8,len(conf['AcceptEnv']))
            keylines = conf.keylines(['AcceptEnv'])
            self.assertEqual(1,len(keylines))
            accept_env_line = keylines[0]
            accept_env_line_array = accept_env_line.split()
            self.assertEqual(3,len(accept_env_line_array))
        return


    def test_target_spaced5(self):
        lines_useradd = self.USERADD_RPMBASED5[::-1].split(linesep)
        with getset.Keyedfile(lines_useradd) as conf:
            self.assertEqual(conf.keysep,chr(61))
            self.assertEqual(conf.linecount,5)
            # Five lines total. Fifth line is the SHELL definition
            val = '[undefined]'
            if 'SHELL' in conf:
                val = conf.keyval('SHELL')
                tup = conf.tuples_all_from_key('SHELL')
            self.assertTrue('/bin/' in val,"value with key 'SHELL' not found or unexpected value")
            if tup is not None:
                """ before_count stored in original tuple (third field) should be 0 """
                self.assertEqual(0,tup[2])
            target_spaced_shell = conf.target_spaced('SHELL')
            self.assertEqual(target_spaced_shell,'^SHELL=')


    def test_target_spaced8(self):
        lines_useradd = self.USERADD_RPMBASED5HACKY[::-1].split(linesep)
        with getset.Keyedfile(lines_useradd) as conf:
            self.assertEqual(conf.keysep,chr(61))
            self.assertEqual(conf.linecount,8)
            # Eight lines total. Fifth, Sixth, Seventh, Eighth lines have some bearing on SHELL definition 
            val = '[undefined]'
            tup = None
            if 'SHELL' in conf:
                val = conf.keyval('SHELL')
                tup = conf.tuples_all_from_key('SHELL')
            self.assertTrue('/bin/' in val,"value with key 'SHELL' not found or unexpected value")
            self.assertEqual('/bin/bash',val)
            if tup is not None:
                """ before_count stored in original tuple (third field) should be 2 """
                self.assertEqual(2,tup[2])
            target_spaced_shell = conf.target_spaced('SHELL')
            #print "TARGET:{0}".format(target_spaced_shell)
            self.assertEqual(target_spaced_shell,'^SHELL  =',"Spacing issue? Line had val={0}".format(val))
        return


    def test_sed_of_target_spaced(self):
        return


    def test_traffic3(self):
        # chr(58) is colon (:) ; chr(61) is equals (=) ; chr(32) is space ; chr(10) is line feed
        with getset.Keyedfile(self.TRAFFIC3OFF_SPLIT) as conf:
            self.assertEqual(conf.keysep,chr(58))
            self.assertEqual(conf.linecount,5)
            conf_items = list(conf.items())
            """ conf is an OrderedDict so the order should be entirely predictable """
            self.assertEqual(conf_items,[('red', 'off'),('orange', 'off'),('green', 'off')])
            self.assertEqual(5,conf.keylen_average)
            self.assertEqual(chr(10),conf.conf_linesep)
        return


    def test_traffic3green(self):
        # chr(58) is colon (:) ; chr(61) is equals (=) ; chr(32) is space ; chr(10) is line feed
        with getset.Keyedfile(self.TRAFFIC3GREEN_SPLIT) as conf:
            self.assertEqual(conf.keysep,chr(58))
            self.assertEqual(conf.linecount,5)
            conf_items = list(conf.items())
            """ conf is an OrderedDict so the order should be entirely predictable """
            """ Writing the following commented line would be wrong as duplicates
            do not make it into the OrderedDict:
            self.assertEqual(conf_items,[('green', 'off'),('green', 'off'),('green', 'on')])
            """
            self.assertEqual(1,len(conf))
            self.assertEqual(2,conf.key_duplicates_count)
            self.assertEqual(5,conf.keylen_average)
            self.assertEqual(chr(10),conf.conf_linesep)
        return


    def test_traffic3hacky(self):
        # chr(58) is colon (:) ; chr(61) is equals (=) ; chr(32) is space ; chr(10) is line feed
        #rint self.TRAFFIC3HACKY_SPLIT
        print "<<< test_traffic3hacky next >>>"
        with getset.Keyedfile(self.TRAFFIC3HACKY_SPLIT) as conf:
            self.assertEqual(conf.keysep,chr(58))
            self.assertEqual(conf.linecount,17)
            self.assertEqual(5,len(conf))
            self.assertEqual(5,conf.keylen_average)
            #print conf.lines_tuple
            #rint conf.tuples_all
            #rint(conf.keylines_disabled())
            enabled = conf.keylines_enabled()
            #print enabled
            enabled = conf.keylines_enabled_resplit()
            #print enabled
            """
            for conf_k,conf_v in conf.items():
                print "{0}DDD{1}".format(conf_k,conf_v)
            for tup in conf.tuples_all:
                # Tuple: key, line, before_count, sep, after_count, linesep, linenum
                print "{0}TTT".format(tup[0])
            """
            """ For the next line we would have 6 if we counted the enabled keys only, but
            when counting duplicates we included keys which begin '#' and count those
            duplicates also. Hence the 10 total
            """
            self.assertEqual(10,conf.key_duplicates_count)
            self.assertEqual(chr(10),conf.conf_linesep)
            orange_commented_count = len(conf.keylines_typed(0,['orange']))
            self.assertEqual(5,orange_commented_count)
            orange_or_commented_orange_count = len(conf.keylines_typed(2,['orange']))
            self.assertEqual(11,orange_or_commented_orange_count)
            """ By loading up on duplicate 'orange' we have managed to boost
            the keylen_average figure from 5 to 6 with calculation shown below:
            To NEAREST INTEGER 6=(4+3+3+3+7+7+7+7+7+6+6+6+6+6+6+5)/16
            We test 6 equality keylen_average_full next """
            self.assertEqual(6,conf.keylen_average_full,"Disagreement: keylen_average_full")
            """ When calculating keylen_average we will have a much shorter list
            of items to sum as we are just interested in keys in the dictionary
            and so duplicates will have no part in the calculation.
            To NEAREST INTEGER 5=(4+3+7+6)/4
            We test 5 equality keylen_average next """
            self.assertEqual(5,conf.keylen_average,"Disagreement: keylen_average")
        return


    def test_traffic3modules(self):
        # chr(58) is colon (:) ; chr(61) is equals (=) ; chr(32) is space ; chr(10) is line feed
        #print self.TRAFFIC3HACKY_SPLIT
        print "<<< traffic3modules detailed next >>>"
        with getset.Keyedfile(self.TRAFFIC3MODULES_SPLIT) as conf:
            self.assertEqual(conf.keysep,chr(58))
            self.assertEqual(conf.linecount,9)
            """
            for conf_k,conf_v in conf.items():
                print "{0}DDD{1}".format(conf_k,conf_v)
            """
            orange_or_commented_orange_count = len(conf.keylines_typed(2,['orange']))
            self.assertEqual(2,orange_or_commented_orange_count)
            self.assertEqual(5,len(conf))
        return


    def test_traffic3inconsistent(self):
        """ Inconsistent as first line looks like following:
        cups server = mycupsserver:1631
        ...but we are a really a colon separated (:) file
        Here keysep is guessed correctly as colon (:)
        """
        # chr(58) is colon (:) ; chr(61) is equals (=) ; chr(32) is space ; chr(10) is line feed
        print "<<< traffic3inconsistent detailed next >>>"
        with getset.Keyedfile(self.TRAFFIC3OFF_BAD_SPLIT) as conf:
            self.assertNotEqual(chr(32),conf.keysep)
            # Test above that keysep is not space, below that keysep is correct (:)
            self.assertEqual(chr(58),conf.keysep)
            self.assertEqual(conf.linecount,6)
            values_list = [ v for v in conf.itervalues() ]
            values_set = set(values_list)
            self.assertEqual(2,len(values_set))
            self.assertEqual(3,len(conf.keylines_typed(1,['red','orange','green'])))
            self.assertTrue(conf.toggle_safe('green','on'))
        return


    def test_traffic3inconsistent2(self):
        """ Inconsistent as first line looks like following:
        mycupsserver 1631
        ...but we are a really a colon separated (:) file
        Here keysep is guessed to be space ( )
        """
        # chr(58) is colon (:) ; chr(61) is equals (=) ; chr(32) is space ; chr(10) is line feed
        print "<<< traffic3inconsistent2 detailed next >>>"
        with getset.Keyedfile(self.TRAFFIC3OFF_BAD2_SPLIT) as conf:
            self.assertNotEqual(chr(58),conf.keysep)
            # Test above that keysep is not space, below that keysep is correct (:)
            self.assertEqual(chr(32),conf.keysep)
            self.assertEqual(conf.linecount,6)
            self.assertEqual(4,len(conf))
            """
            for conf_k,conf_v in conf.items():
                print "{0}DDD{1}".format(conf_k,conf_v)
            """
            values_list = [ v for v in conf.itervalues() ]
            values_set_nonblanks = set([ v for v in values_list if len(v) > 0])
            self.assertEqual(1,len(values_set_nonblanks))
            """ Should just be single non blank value of 1631 """
            self.assertEqual(0,len(conf.keylines_typed(1,['red','orange','green'])))
            self.assertFalse(conf.toggle_safe('green','on'))
        return


    def test_traffic3inconsistent3(self):
        """ Inconsistent as first line looks like following:
        port        = 3306
        ...but we are a really a colon separated (:) file
        Here keysep is guessed to be equals (=)
        """
        # chr(58) is colon (:) ; chr(61) is equals (=) ; chr(32) is space ; chr(10) is line feed
        print "<<< traffic3inconsistent3 detailed next >>>"
        with getset.Keyedfile(self.TRAFFIC3OFF_BAD3_SPLIT) as conf:
            self.assertNotEqual(chr(58),conf.keysep)
            # Test above that keysep is not space, below that keysep is correct (:)
            self.assertEqual(chr(61),conf.keysep)
            self.assertEqual(conf.linecount,6)
            """
            for conf_k,conf_v in conf.items():
                print "{0}DDD{1}".format(conf_k,conf_v)
            """
            values_list = [ v for v in conf.itervalues() ]
            values_nonblanks = [ v for v in values_list if len(v) > 0]
            self.assertEqual(1,len(values_nonblanks))
            """ Should just be single non blank value of 3306 """
            self.assertEqual('3306',values_nonblanks[0])
            # Final two parameters of next command says safe=True and autoquote_level=1
            conf.toggle_safe('green','on',True,1)
        return


    def test_toggle_twice_disallowed(self):
        lines_useradd = self.CRON_READ_ENV[::-1].split(linesep)
        with getset.Keyedfile(lines_useradd) as conf:
            self.assertEqual(conf.keysep,chr(61))
            self.assertEqual(conf.linecount,4)
            self.assertFalse(conf.toggle_safe('ENABLED','true',True))
            """ Above should fail as we do not have an ENABLED key """
            self.assertFalse(conf.toggle_safe('READ_ENV',"'no'",True))
            """ Above should fail because when quotes are provided in
            the new value, and third arg is True, there is a safety
            check to ensure we are not adding quotes that were not there
            around the existing value. If this was successful then
            our program just allowed us to switch the quotes from double (")
            around the original value, to single (') around our new value.
            """
            self.assertFalse(conf.toggle_safe('READ_ENV','no',True))
            self.assertTrue(conf.toggle_safe('READ_ENV','no',True,1))
            """ Trying the same thing twice should fail """
            self.assertFalse(conf.toggle_safe('READ_ENV','no',True,1))
        return


    def test_lines_for_put_simple(self):
        lines_useradd = self.CRON_READ_ENV[::-1].split(linesep)
        with getset.Keyedfile(lines_useradd) as conf:
            self.assertEqual(conf.keysep,chr(61))
            self.assertEqual(conf.linecount,4)
            self.assertTrue(conf.toggle_safe('READ_ENV','no',True,1))
            put_lines = conf.lines_for_put()
            read_env_count = 0
            for line in put_lines:
                if 'READ_ENV' in line:
                    read_env_count += 1
            self.assertEqual(1,read_env_count)
            self.assertEqual(4,len(put_lines))
        return


    def test_linecount_max(self):
        """ conf_get_set does not really support ini style files with sections,
        there are already better solutions (configparser) for handling them.
        However for testing and examples the large file length and other features
        of php.ini can provide useful input
        """
        # chr(33) is exclamation mark (!)      ; chr(34) is double quote (")     ; chr(39) is single quote (')
        # chr(35) is hash / octothorpe             ; chr(58) is colon (:)            ; chr(61) is equals (=)
        # chr(91) is open bracket ([)      ; chr(93) is close bracket (])
        # chr(59) is semicolon (;)             ; chr(45) is minus sign  (-)

        #print self.PHP5INI_SIMILAR

        similar_len = len(self.PHP5INI_SIMILAR_SPLIT)
        print "Attempt to open a conf (Keyedfile) having {0} lines next".format(similar_len)
        space_equals_space_count = 0
        with getset.Keyedfile(self.PHP5INI_SIMILAR_SPLIT) as conf:
            self.assertTrue(set(conf.keysep).issubset(self.SET_COLON_SPACE_EQUALS))
            for tup in conf.tuples_all:
                # Tuple: key, line, before_count, sep, after_count, linesep, linenum
                if tup[2] == 1 and tup[3] == chr(61) and tup[4] == 1:
                    space_equals_space_count += 1
                #print tup
        """ There are 3 lines of the form key = value in PHP5INI_SIMILAR
        and they are consistently spaced. But we also deliberately added
        an equals into a comment line. We do not recognise semicolon as
        meaning a comment line and so split that line as if it were a key value
        and so that makes 4 """
        self.assertEqual(4,space_equals_space_count)
                    
        large_len = len(self.PHP5INI_SIMILAR_LARGE_SPLIT)
        print "Attempt to open a conf (Keyedlarge) having {0} lines next".format(large_len)
        with getset.Keyedlarge(self.PHP5INI_SIMILAR_LARGE_SPLIT) as conf:
            self.assertEqual(conf.keysep,chr(61))
            self.assertEqual(conf.linecount,2021)

        print "Attempt to open a conf (Keyedfile) having {0} lines next".format(large_len)
        """ All working as expected we should see an Exception thrown by Keyedfile, which
        is caught and examined to ensure it contains 'MAX_LINES limit exceeded'.
        If instead we encounter an AssertionError or an Exception that does not contain
        our expected string, then we should raise or report something
        """
        from exceptions import AssertionError
        try:
            with getset.Keyedfile(self.PHP5INI_SIMILAR_LARGE_SPLIT) as conf:
                """ We should not reach this far as Exception should have been thrown """
                self.assertEqual(conf.keysep,chr(61))
        except AssertionError:
            raise
        except Exception,ex:
            """ Expect Exception to have been thrown
            and assert that it is correct Exception """
            self.assertTrue('MAX_LINES limit exceeded' in str(ex))
        return
    

    def test_repeated_and_single_keys(self):
        """ Here we test input which has repeated keys. Normally this would happen if
        conf_get_set was being used against a file which was the subject of many
        manual alterations, or a file which has a nested structure.
        conf_get_set is not designed to deal with sections or nesting, but there is no harm
        in defining expected behaviour tests and it may provide a useful illustration / example
        """
        # chr(58) is colon (:)            ; chr(61) is equals (=)           ; chr(32) is space
        """ Now what happens for our sample ports.conf is that the first 6 lines are treated
        well, and most things after those 6 lines produce a duplicate key. This 7 item list may help
          ['<IfModule', '#If', '#the', '#to', 'Listen', '</IfModule>', '#Listen']
        So whilst it is perfectly normal to have repeated directives in apache, we are treating
        the ports.conf file as a flat key value structure. In our case keys which are encountered
        again are treated as a redefinition. Do read the traffic3hacky tests also.
        """
        with getset.Keyedfile(self.PORTS2SECTION_SPLIT) as conf:
            #print conf.tuples_all
            self.assertEqual(conf.keysep,chr(32))
            """ Above checks that by the end of file the separator between key and value
            is thought to be space ' ' """
            self.assertEqual(conf.linecount,13)
            """ So we have 13 lines of which 7 are unique keys """
            self.assertEqual(7,len(conf))
            self.assertEqual(3,conf.key_duplicates_count)
            """ If you consider enabled as 'not beginning with #' then you will expect 6 next """
            self.assertEqual(6,len(conf.tuples_enabled()))
            """ All non-blank lines were marked with keysep ' ' during initial analysis so ...."""
            self.assertEqual(0,len(conf.tuples_without_keysep()))
            self.assertEqual(6,len(conf.tuples_with_keysep()))
            self.assertEqual(6,len(conf.tuples_with_keysep(chr(32))))
            """ Listen was orinally set to 443 but in the second <IfModule block
            it is redefined as 44443. """
            self.assertEqual('44443',conf['Listen'])
        return


    def test_toggle_success(self):
        from hashlib import md5
        mdf = None
        print "<<< test_toggle_success next >>>"
        with getset.Keyedfile(self.MYSQL5CNF_SIMILAR_SPLIT) as conf:
            self.assertEqual(conf.keysep,chr(61))
            self.assertEqual(23,conf.linecount)
            self.assertEqual(9,conf.key_duplicates_count)
            mdf = md5()
            for line in conf.lines_tuple:
                #print line
                mdf.update(line)
            self.assertIn('skip-networking',conf)
            self.assertTrue(conf.toggle_safe('skip-networking','off'))
            self.assertEqual(1,len(conf.dict_of_seds))
            #for sed in conf.list_of_sedlines: print sed
            self.assertIn(21,conf.dict_of_seds)
            # Above checks that line 21 was identified by our toggle_safe call as the target 
        self.assertNotEqual(None,mdf)
        self.assertEqual('5d2f0aae1c84592bd1289e0fec4daa6b',mdf.hexdigest())
        return


    def test_toggle_success2(self):
        print "<<< test_toggle_success2 next >>>"
        with getset.Keyedfile(self.MERGETOOLSRC_SPLIT) as conf:
            self.assertEqual(conf.keysep,chr(61))
            self.assertEqual(13,conf.linecount)
            self.assertEqual(0,conf.key_duplicates_count)
            self.assertEqual(11,len(conf))
            """ If you consider enabled as 'not beginning with #' then you will expect 6 next """
            self.assertEqual(11,len(conf.tuples_enabled()))
            self.assertIn('meld.gui',conf)
            self.assertTrue(conf.toggle_safe('meld.gui','False'))
            self.assertTrue(conf.toggle_safe('tkdiff.gui','false'))
        return


    def test_toggle_fail1(self):
        print "<<< test_toggle_fail1 next >>>"
        with getset.Keyedfile(self.MYSQL5CNF_SIMILAR_SPLIT) as conf:
            self.assertEqual(conf.keysep,chr(61))
            self.assertEqual(23,conf.linecount)
            self.assertEqual(9,conf.key_duplicates_count)
            self.assertGreaterEqual(2,len(conf['skip-networking']),"Expected on or off for skip-networking")
            """ Next lines_enabled_contains() consults the original line
            as stored in the 7 tuple at creation time """
            self.assertEqual(3,len(conf.lines_enabled_contains('skip-networking')))
            self.assertEqual(['  skip-networking   = on',
                              '   skip-networking   = off',
                              '  skip-networking   = on'],
                             conf.lines_enabled_contains('skip-networking'))
            # Above test assertEqual will automatically use type specific assertListEqual
            """ Next we expect failure as toggle_safe() is supposed to 
            fail when our toggle would result in
            no change in on/off state """
            self.assertFalse(conf.toggle_safe('skip-networking','on'))
        return


    def test_toggle_and_put1(self):
        """ In this test we go as far as checking that lines_for_put() returns
        the expected representation of the file lines after successful toggle
        """
        print "<<< test_toggle_and_put1 next >>>"
        with getset.Keyedfile(self.MYSQL5CNF_SIMILAR_SPLIT) as conf:
            self.assertEqual(conf.keysep,chr(61))
            self.assertEqual(23,conf.linecount)
            self.assertTrue(conf.toggle_safe('skip-networking','off'))
            md5before = '5d2f0aae1c84592bd1289e0fec4daa6b'
            self.assertEqual(md5before,cmdbin.md5_from_iterable(conf.lines_tuple))
            md5after = '9c7d92f15bdceb21640fe98007e7bae9'
            self.assertEqual(md5after,cmdbin.md5_from_iterable(conf.lines_for_put()))
        return


    def test_toggle_and_put3(self):
        """ In this test we go as far as checking that lines_for_put() returns
        the expected representation of the file lines after successful toggle
        """
        print "<<< test_toggle_and_put3 next >>>"
        with getset.Keyedfile(self.MERGETOOLSRC_SPLIT) as conf:
            self.assertEqual(conf.keysep,chr(61))
            self.assertEqual(13,conf.linecount)
            self.assertEqual(0,conf.key_duplicates_count)
            self.assertEqual(11,len(conf))
            self.assertIn('meld.gui',conf)
            self.assertTrue(conf.toggle_safe('meld.gui','False'))
            self.assertTrue(conf.toggle_safe('tkdiff.gui','false'))
            self.assertTrue(conf.toggle_safe('xxdiff.gui','false'))
            md5before = 'f1f7a423dd0a80caf39daaa19f69072d'
            self.assertEqual(md5before,cmdbin.md5_from_iterable(conf.lines_tuple))
            md5after = '1835f2ddfcbbcf47b2e526dd789c526a'
            self.assertEqual(md5after,cmdbin.md5_from_iterable(conf.lines_for_put()))
        return

   
    def test_toggle_and_sedlines1(self):
        # chr(58) is colon (:) ; chr(61) is equals (=) ; chr(32) is space ; chr(10) is line feed
        print "<<< test_toggle_and_sedlines1 next >>>"
        with getset.Keyedfile(self.MYSQL5CNF_SIMILAR_SPLIT) as conf:
            self.assertEqual(conf.keysep,chr(61))
            self.assertEqual(23,conf.linecount)
            self.assertIn('skip-networking',conf)
            self.assertTrue(conf.toggle_safe('skip-networking','off'))
            self.assertEqual(1,len(conf.dict_of_seds))
            SKIPNET = 'skip-networking'
            sedline = "0,/^  {0}   = / s/\(^  {0}   = \)\(.*\)/\\1off/".format(SKIPNET)
            # Above we use \\1 below when really we mean \1 so as Python expresses it properly
            self.assertEqual(conf.dict_of_seds[21],sedline)
            #for linenum,sed in conf.dict_of_seds.iteritems():
            #    print "{0}    {1}".format(linenum,sed)
            scriptlines,md5after,lines = conf.sed_checksum(conf.list_of_sedlines,
                                                           debug_filename='toggle_and_sedlines1')
            self.assertIsNotNone(md5after)
            textreport,compared = cmdbin.iterable1iterable2compare(conf.lines_tuple,lines,False)
            self.assertEqual(md5after,'9c7d92f15bdceb21640fe98007e7bae9')
            self.assertFalse(any(compared))
            self.assertGreater(scriptlines,1)
            self.assertGreater(lines,1)
        return


    def test_value_diacritic(self):
        print "<<< test_value_diacritic next >>>"
        with getset.Keyedfile(self.MERGETOOLSDIACRITIC_SPLIT) as conf:
            self.assertEqual(conf.keysep,chr(61))
            self.assertEqual(12,conf.linecount)
            self.assertIn('ecmerge.regkey',conf)
            regkey = conf['ecmerge.regkey']
            regkey8 = regkey.decode('utf-8')
            unichr233 = unichr(233)
            #print regkey.decode('utf-8')
            self.assertIn(unichr(233),regkey8)
            target_name = 'LATIN SMALL LETTER E WITH ACUTE'
            self.assertGreaterEqual(cmdbin.unicode_name_index(target_name,regkey8),0)
            self.assertGreater(cmdbin.unicode_name_index(target_name,regkey8[:5]),3)
        return


    def test_one_liners(self):
        """
        Test the following one liners to see if guesswork alone gives the
        correct delimiter (space) in all cases.
        self.PORTS1LISTEN = "".join(reversed('Listen 80{0}'.format(linesep)))
        self.PORTS1LISTENAST = "".join(reversed('Listen *:80{0}'.format(linesep)))
        self.PORTS1LISTENLOCAL = "".join(reversed('Listen 127.0.0.1:80{0}'.format(linesep)))
        """
        # chr(58) is colon (:) ; chr(61) is equals (=) ; chr(32) is space ; chr(10) is line feed
        lines = self.PORTS1LISTENAST[::-1].split(linesep)
        with getset.Keyedspace(lines) as conf:
            self.assertEqual(conf.keysep,chr(32))
            self.assertEqual('*:80',conf['Listen'])
        lines = self.PORTS1LISTENLOCAL[::-1].split(linesep)
        with getset.Keyedspace(lines) as conf:
            self.assertEqual(conf.keysep,chr(32))
            self.assertEqual('127.0.0.1:80',conf['Listen'])
        lines = self.PORTS1LISTEN[::-1].split(linesep)
        with getset.Keyedspace(lines) as conf:
            self.assertEqual(conf.keysep,chr(32))
            self.assertEqual('80',conf['Listen'])
        """ Having tested things with the specialist class Keyedspace,
        now use the generic (guess the separator) class Keyedfile
        """
        lines = self.PORTS1LISTENAST[::-1].split(linesep)
        with getset.Keyedfile(lines) as conf:
            self.assertEqual(conf.keysep,chr(32))
            self.assertEqual('*:80',conf['Listen'])
        lines = self.PORTS1LISTENLOCAL[::-1].split(linesep)
        with getset.Keyedfile(lines) as conf:
            self.assertEqual(conf.keysep,chr(32))
            self.assertEqual('127.0.0.1:80',conf['Listen'])
        lines = self.PORTS1LISTEN[::-1].split(linesep)
        with getset.Keyedfile(lines) as conf:
            self.assertEqual(conf.keysep,chr(32))
            self.assertEqual('80',conf['Listen'])
